# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

"""
Access to SPI Flash parts

usage:
    >>> read_spi( spi_fla, length )
    >>> write_spi( spi_fla, buf )
    >>> erase_spi_block( spi_fla )
    >>> get_SPI_JEDEC_ID()
    >>> get_SPI_JEDEC_ID_decoded()

.. note::
    !! IMPORTANT:
    Size of the data chunk used in SPI read cycle (in bytes)
    default = maximum 64 bytes (remainder is read in 4 byte chunks)

    If you want to change logic to read SPI Flash in 4 byte chunks:
    SPI_READ_WRITE_MAX_DBC = 4

    @TBD: SPI write cycles operate on 4 byte chunks (not optimized yet)

    Approximate performance (on 2-core SMT Intel Core i5-4300U (Haswell) CPU 1.9GHz):
    SPI read: ~7 sec per 1MB (with DBC=64)
"""

import struct
import time
from typing import Dict, Tuple, Optional
from chipsec.library.defines import ALIGNED_4KB, BIT0, BIT1, BIT2, BIT5
from chipsec.library.file import write_file, read_file
from chipsec.library.logger import print_buffer_bytes
from chipsec.hal import hal_base, mmio
from chipsec.hal.spi_jedec_ids import JEDEC_ID
from chipsec.library.exceptions import SpiRuntimeError, UnimplementedAPIError

SPI_READ_WRITE_MAX_DBC = 64
SPI_READ_WRITE_DEF_DBC = 4
SFDP_HEADER = 0x50444653

SPI_MAX_PR_COUNT = 5
SPI_FLA_SHIFT = 12
SPI_FLA_PAGE_MASK = ALIGNED_4KB

SPI_MMIO_BASE_LENGTH = 0x200
PCH_RCBA_SPI_HSFSTS_SCIP = BIT5                          # SPI cycle in progress
PCH_RCBA_SPI_HSFSTS_AEL = BIT2                          # Access Error Log
PCH_RCBA_SPI_HSFSTS_FCERR = BIT1                          # Flash Cycle Error
PCH_RCBA_SPI_HSFSTS_FDONE = BIT0                          # Flash Cycle Done

PCH_RCBA_SPI_HSFCTL_FCYCLE_READ = 0                             # Flash Cycle Read
PCH_RCBA_SPI_HSFCTL_FCYCLE_WRITE = 2                             # Flash Cycle Write
PCH_RCBA_SPI_HSFCTL_FCYCLE_ERASE = 3                             # Flash Cycle Block Erase
PCH_RCBA_SPI_HSFCTL_FCYCLE_SFDP = 5
PCH_RCBA_SPI_HSFCTL_FCYCLE_JEDEC = 6                             # Flash Cycle Read JEDEC ID
PCH_RCBA_SPI_HSFCTL_FCYCLE_FGO = BIT0                          # Flash Cycle GO

PCH_RCBA_SPI_FADDR_MASK = 0x07FFFFFF                      # SPI Flash Address Mask [0:26]

PCH_RCBA_SPI_FREGx_LIMIT_MASK = 0x7FFF0000                    # Size
PCH_RCBA_SPI_FREGx_BASE_MASK = 0x00007FFF                    # Base

PCH_RCBA_SPI_OPTYPE_RDNOADDR = 0x00
PCH_RCBA_SPI_OPTYPE_WRNOADDR = 0x01
PCH_RCBA_SPI_OPTYPE_RDADDR = 0x02
PCH_RCBA_SPI_OPTYPE_WRADDR = 0x03

PCH_RCBA_SPI_FDOC_FDSS_FSDM = 0x0000                        # Flash Signature and Descriptor Map
PCH_RCBA_SPI_FDOC_FDSS_COMP = 0x1000                        # Component
PCH_RCBA_SPI_FDOC_FDSS_REGN = 0x2000                        # Region
PCH_RCBA_SPI_FDOC_FDSS_MSTR = 0x3000                        # Master
PCH_RCBA_SPI_FDOC_FDSI_MASK = 0x0FFC                        # Flash Descriptor Section Index

# agregated SPI Flash commands
HSFCTL_READ_CYCLE = ((PCH_RCBA_SPI_HSFCTL_FCYCLE_READ << 1) | PCH_RCBA_SPI_HSFCTL_FCYCLE_FGO)
HSFCTL_WRITE_CYCLE = ((PCH_RCBA_SPI_HSFCTL_FCYCLE_WRITE << 1) | PCH_RCBA_SPI_HSFCTL_FCYCLE_FGO)
HSFCTL_ERASE_CYCLE = ((PCH_RCBA_SPI_HSFCTL_FCYCLE_ERASE << 1) | PCH_RCBA_SPI_HSFCTL_FCYCLE_FGO)
HSFCTL_JEDEC_CYCLE = ((PCH_RCBA_SPI_HSFCTL_FCYCLE_JEDEC << 1) | PCH_RCBA_SPI_HSFCTL_FCYCLE_FGO)
HSFCTL_SFDP_CYCLE = ((PCH_RCBA_SPI_HSFCTL_FCYCLE_SFDP << 1) | PCH_RCBA_SPI_HSFCTL_FCYCLE_FGO)

# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# FGO bit cleared (for safety ;)
# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# HSFCTL_WRITE_CYCLE = ( (PCH_RCBA_SPI_HSFCTL_FCYCLE_WRITE<<1) )
# HSFCTL_ERASE_CYCLE = ( (PCH_RCBA_SPI_HSFCTL_FCYCLE_ERASE<<1) )

HSFSTS_CLEAR = (PCH_RCBA_SPI_HSFSTS_AEL | PCH_RCBA_SPI_HSFSTS_FCERR | PCH_RCBA_SPI_HSFSTS_FDONE)

#
# Hardware Sequencing Flash Status (HSFSTS)
#
SPI_HSFSTS_OFFSET = 0x04
# HSFSTS bit masks
SPI_HSFSTS_FLOCKDN_MASK = (1 << 15)
SPI_HSFSTS_FDOPSS_MASK = (1 << 13)

#
# Flash Regions
#

SPI_REGION_NUMBER_IN_FD = 12

FLASH_DESCRIPTOR = 0
BIOS = 1
ME = 2
GBE = 3
PLATFORM_DATA = 4
FREG5 = 5
FREG6 = 6
FREG7 = 7
EMBEDDED_CONTROLLER = 8
FREG9 = 9
FREG10 = 10
FREG11 = 11

SPI_REGION: Dict[int, str] = {
    FLASH_DESCRIPTOR: 'FREG0_FLASHD',
    BIOS: 'FREG1_BIOS',
    ME: 'FREG2_ME',
    GBE: 'FREG3_GBE',
    PLATFORM_DATA: 'FREG4_PD',
    FREG5: 'FREG5',
    FREG6: 'FREG6',
    FREG7: 'FREG7',
    EMBEDDED_CONTROLLER: 'FREG8_EC',
    FREG9: 'FREG9',
    FREG10: 'FREG10',
    FREG11: 'FREG11'
}

SPI_REGION_NAMES: Dict[int, str] = {
    FLASH_DESCRIPTOR: 'Flash Descriptor',
    BIOS: 'BIOS',
    ME: 'Intel ME',
    GBE: 'GBe',
    PLATFORM_DATA: 'Platform Data',
    FREG5: 'Flash Region 5',
    FREG6: 'Flash Region 6',
    FREG7: 'Flash Region 7',
    EMBEDDED_CONTROLLER: 'Embedded Controller',
    FREG9: 'Flash Region 9',
    FREG10: 'Flash Region 10',
    FREG11: 'Flash Region 11'
}

#
# Flash Descriptor Master Defines
#

MASTER_HOST_CPU_BIOS = 0
MASTER_ME = 1
MASTER_GBE = 2
MASTER_EC = 3

SPI_MASTER_NAMES: Dict[int, str] = {
    MASTER_HOST_CPU_BIOS: 'CPU',
    MASTER_ME: 'ME',
    MASTER_GBE: 'GBe',
    MASTER_EC: 'EC'
}

DEFAULT_PROGRESS_MAX = 50

# @TODO: DEPRECATED


def get_SPI_region(flreg: int) -> Tuple[int, int]:
    range_base = (flreg & PCH_RCBA_SPI_FREGx_BASE_MASK) << SPI_FLA_SHIFT
    range_limit = ((flreg & PCH_RCBA_SPI_FREGx_LIMIT_MASK) >> 4)
    range_limit |= SPI_FLA_PAGE_MASK
    return (range_base, range_limit)


class SPI(hal_base.HALBase):

    def __init__(self, cs):
        super(SPI, self).__init__(cs)
        self.mmio = mmio.MMIO(cs)
        self.rcba_spi_base = self.get_SPI_MMIO_base()
        # We try to map SPIBAR in the process memory, this will increase the
        # speed of MMIO access later on.
        try:
            self.cs.helper.map_io_space(self.rcba_spi_base, SPI_MMIO_BASE_LENGTH, 0)
        except UnimplementedAPIError:
            pass

        # Reading definitions of SPI flash controller registers
        # which are required to send SPI cycles once for performance reasons
        self.hsfs_off = self.cs.register.get_def("HSFS")['offset']
        self.hsfc_off = self.cs.register.get_def("HSFC")['offset']
        self.faddr_off = self.cs.register.get_def("FADDR")['offset']
        self.fdata0_off = self.cs.register.get_def("FDATA0")['offset']
        self.fdata1_off = self.cs.register.get_def("FDATA1")['offset']
        self.fdata2_off = self.cs.register.get_def("FDATA2")['offset']
        self.fdata3_off = self.cs.register.get_def("FDATA3")['offset']
        self.fdata4_off = self.cs.register.get_def("FDATA4")['offset']
        self.fdata5_off = self.cs.register.get_def("FDATA5")['offset']
        self.fdata6_off = self.cs.register.get_def("FDATA6")['offset']
        self.fdata7_off = self.cs.register.get_def("FDATA7")['offset']
        self.fdata8_off = self.cs.register.get_def("FDATA8")['offset']
        self.fdata9_off = self.cs.register.get_def("FDATA9")['offset']
        self.fdata10_off = self.cs.register.get_def("FDATA10")['offset']
        self.fdata11_off = self.cs.register.get_def("FDATA11")['offset']
        self.fdata12_off = self.cs.register.get_def("FDATA12")['offset']
        self.fdata13_off = self.cs.register.get_def("FDATA13")['offset']
        self.fdata14_off = self.cs.register.get_def("FDATA14")['offset']
        self.fdata15_off = self.cs.register.get_def("FDATA15")['offset']
        self.bios_ptinx = self.cs.register.get_def("BIOS_PTINX")['offset']
        self.bios_ptdata = self.cs.register.get_def("BIOS_PTDATA")['offset']

        self.logger.log_hal("[spi] Reading SPI flash controller registers definitions:")
        self.logger.log_hal(f'      HSFS   offset = 0x{self.hsfs_off:04X}')
        self.logger.log_hal(f'      HSFC   offset = 0x{self.hsfc_off:04X}')
        self.logger.log_hal(f'      FADDR  offset = 0x{self.faddr_off:04X}')
        self.logger.log_hal(f'      FDATA0 offset = 0x{self.fdata0_off:04X}')

    def get_SPI_MMIO_base(self) -> int:
        spi_base = 0
        if self.mmio.is_MMIO_BAR_defined('SPIBAR'):
            (spi_base, _) = self.mmio.get_MMIO_BAR_base_address('SPIBAR')
        else:
            self.logger.log_hal('[spi] get_SPI_MMIO_base(): SPIBAR not defined. Returning spi_base = 0.')
        self.logger.log_hal(f'[spi] SPI MMIO base: 0x{spi_base:016X} (assuming below 4GB)')
        return spi_base

    def spi_reg_read(self, reg: int, size: int = 4) -> int:
        return self.mmio.read_MMIO_reg(self.rcba_spi_base, reg, size)

    def spi_reg_write(self, reg: int, value: int, size: int = 4) -> Optional[int]:
        return self.mmio.write_MMIO_reg(self.rcba_spi_base, reg, value, size)

    def get_SPI_region(self, spi_region_id: int) -> Tuple[int, int, int]:
        freg_name = SPI_REGION[spi_region_id]
        if not self.cs.register.is_defined(freg_name):
            return (0, 0, 0)
        freg = self.cs.register.read(freg_name)
        # Region Base corresponds to FLA bits 24:12
        range_base = self.cs.register.get_field(freg_name, freg, 'RB') << SPI_FLA_SHIFT
        # Region Limit corresponds to FLA bits 24:12
        range_limit = self.cs.register.get_field(freg_name, freg, 'RL') << SPI_FLA_SHIFT
        # FLA bits 11:0 are assumed to be FFFh for the limit comparison
        range_limit |= SPI_FLA_PAGE_MASK
        return (range_base, range_limit, freg)

    SpiRegions = Dict[int, Tuple[int, int, int, str, int]]

    # all_regions = True : return all SPI regions
    # all_regions = False: return only available SPI regions (limit >= base)
    def get_SPI_regions(self, all_regions: bool = True) -> SpiRegions:
        spi_regions: Dict[int, Tuple[int, int, int, str, int]] = {}
        for r in SPI_REGION:
            (range_base, range_limit, freg) = self.get_SPI_region(r)
            if range_base is None:
                continue
            if all_regions or (range_limit >= range_base):
                range_size = range_limit - range_base + 1
                spi_regions[r] = (range_base, range_limit, range_size, SPI_REGION_NAMES[r], freg)
        return spi_regions

    def get_SPI_Protected_Range(self, pr_num: int) -> Tuple[int, int, int, int, int, int]:
        if pr_num > SPI_MAX_PR_COUNT:
            return (0, 0, 0, 0, 0, 0)

        pr_name = f'PR{pr_num:x}'
        pr_j_reg = self.cs.register.get_def(pr_name)['offset']
        pr_j = self.cs.register.read(pr_name)

        # Protected Range Base corresponds to FLA bits 24:12
        base = self.cs.register.get_field(pr_name, pr_j, 'PRB') << SPI_FLA_SHIFT
        # Protected Range Limit corresponds to FLA bits 24:12
        limit = self.cs.register.get_field(pr_name, pr_j, 'PRL') << SPI_FLA_SHIFT

        wpe = (0 != self.cs.register.get_field(pr_name, pr_j, 'WPE'))
        rpe = (0 != self.cs.register.get_field(pr_name, pr_j, 'RPE'))

        # Check if this is a valid PRx config
        if wpe or rpe:
            # FLA bits 11:0 are assumed to be FFFh for the limit comparison
            limit |= SPI_FLA_PAGE_MASK

        return (base, limit, wpe, rpe, pr_j_reg, pr_j)

    ##############################################################################################################
    # SPI configuration
    ##############################################################################################################

    def display_SPI_Flash_Descriptor(self) -> None:
        self.logger.log("============================================================")
        self.logger.log("SPI Flash Descriptor")
        self.logger.log("------------------------------------------------------------")
        self.logger.log("\nFlash Signature and Descriptor Map:")
        for j in range(5):
            self.cs.register.write('FDOC', (PCH_RCBA_SPI_FDOC_FDSS_FSDM | (j << 2)))
            fdod = self.cs.register.read('FDOD')
            self.logger.log(f'{fdod:08X}')

        self.logger.log("\nComponents:")
        for j in range(3):
            self.cs.register.write('FDOC', (PCH_RCBA_SPI_FDOC_FDSS_COMP | (j << 2)))
            fdod = self.cs.register.read('FDOD')
            self.logger.log(f'{fdod:08X}')

        self.logger.log("\nRegions:")
        for j in range(5):
            self.cs.register.write('FDOC', (PCH_RCBA_SPI_FDOC_FDSS_REGN | (j << 2)))
            fdod = self.cs.register.read('FDOD')
            self.logger.log(f'{fdod:08X}')

        self.logger.log("\nMasters:")
        for j in range(3):
            self.cs.register.write('FDOC', (PCH_RCBA_SPI_FDOC_FDSS_MSTR | (j << 2)))
            fdod = self.cs.register.read('FDOD')
            self.logger.log(f'{fdod:08X}')

    def display_SPI_opcode_info(self) -> None:
        self.logger.log("============================================================")
        self.logger.log("SPI Opcode Info")
        self.logger.log("------------------------------------------------------------")
        preop = self.cs.register.read('PREOP')
        self.logger.log(f'PREOP : 0x{preop:04X}')
        optype = self.cs.register.read('OPTYPE')
        self.logger.log(f'OPTYPE: 0x{optype:04X}')
        opmenu_lo = self.cs.register.read('OPMENU_LO')
        opmenu_hi = self.cs.register.read('OPMENU_HI')
        opmenu = ((opmenu_hi << 32) | opmenu_lo)
        self.logger.log(f'OPMENU: 0x{opmenu:016X}')
        self.logger.log('')
        preop0 = preop & 0xFF
        preop1 = (preop >> 8) & 0xFF
        self.logger.log(f'Prefix Opcode 0 = 0x{preop0:02X}')
        self.logger.log(f'Prefix Opcode 1 = 0x{preop1:02X}')

        self.logger.log("------------------------------------------------------------")
        self.logger.log("Opcode # | Opcode | Optype | Description")
        self.logger.log("------------------------------------------------------------")
        for j in range(8):
            optype_j = ((optype >> j * 2) & 0x3)
            if (PCH_RCBA_SPI_OPTYPE_RDNOADDR == optype_j):
                desc = 'SPI read cycle without address'
            elif (PCH_RCBA_SPI_OPTYPE_WRNOADDR == optype_j):
                desc = 'SPI write cycle without address'
            elif (PCH_RCBA_SPI_OPTYPE_RDADDR == optype_j):
                desc = 'SPI read cycle with address'
            elif (PCH_RCBA_SPI_OPTYPE_WRADDR == optype_j):
                desc = 'SPI write cycle with address'
            else:
                desc = ''
            self.logger.log(f'Opcode{j:d}  | 0x{(opmenu >> j * 8) & 0xFF:02X}   | {optype_j:x}      | {desc} ')

    def display_SPI_Flash_Regions(self) -> None:
        self.logger.log("------------------------------------------------------------")
        self.logger.log("Flash Region             | FREGx Reg | Base     | Limit     ")
        self.logger.log("------------------------------------------------------------")
        regions = self.get_SPI_regions()
        for (region_id, region) in regions.items():
            base, limit, size, name, freg = region
            self.logger.log(f'{region_id:d} {name:22} | {freg:08X}  | {base:08X} | {limit:08X} ')

    def display_BIOS_region(self) -> None:
        bfpreg = self.cs.register.read('BFPR')
        base = self.cs.register.get_field('BFPR', bfpreg, 'PRB') << SPI_FLA_SHIFT
        limit = self.cs.register.get_field('BFPR', bfpreg, 'PRL') << SPI_FLA_SHIFT
        limit |= SPI_FLA_PAGE_MASK
        self.logger.log("BIOS Flash Primary Region")
        self.logger.log("------------------------------------------------------------")
        self.logger.log(f'BFPREG = {bfpreg:08X}:')
        self.logger.log(f'  Base  : {base:08X}')
        self.logger.log(f'  Limit : {limit:08X}')

    def display_SPI_Ranges_Access_Permissions(self) -> None:
        self.logger.log("SPI Flash Region Access Permissions")
        self.logger.log("------------------------------------------------------------")
        fracc = self.cs.register.read('FRAP')
        if self.logger.HAL:
            self.cs.register.print('FRAP', fracc)
        brra = self.cs.register.get_field('FRAP', fracc, 'BRRA')
        brwa = self.cs.register.get_field('FRAP', fracc, 'BRWA')
        bmrag = self.cs.register.get_field('FRAP', fracc, 'BMRAG')
        bmwag = self.cs.register.get_field('FRAP', fracc, 'BMWAG')
        self.logger.log('')
        self.logger.log(f'BIOS Region Write Access Grant ({bmwag:02X}):')
        regions = self.get_SPI_regions()
        for region_id in regions:
            self.logger.log(f'  {SPI_REGION[region_id]:12}: {0 != bmwag & (1 << region_id):1d}')
        self.logger.log(f'BIOS Region Read Access Grant ({bmrag:02X}):')
        for region_id in regions:
            self.logger.log(f'  {SPI_REGION[region_id]:12}: {0 != bmrag & (1 << region_id):1d}')
        self.logger.log(f'BIOS Region Write Access ({brwa:02X}):')
        for region_id in regions:
            self.logger.log(f'  {SPI_REGION[region_id]:12}: {0 != brwa & (1 << region_id):1d}')
        self.logger.log(f'BIOS Region Read Access ({brra:02X}):')
        for region_id in regions:
            self.logger.log(f'  {SPI_REGION[region_id]:12}: {0 != brra & (1 << region_id):1d}')

    def display_SPI_Protected_Ranges(self) -> None:
        self.logger.log("SPI Protected Ranges")
        self.logger.log("------------------------------------------------------------")
        self.logger.log("PRx (offset) | Value    | Base     | Limit    | WP? | RP?")
        self.logger.log("------------------------------------------------------------")
        for j in range(5):
            (base, limit, wpe, rpe, pr_reg_off, pr_reg_value) = self.get_SPI_Protected_Range(j)
            self.logger.log(f'PR{j:d} ({pr_reg_off:02X})     | {pr_reg_value:08X} | {base:08X} | {limit:08X} | {wpe:d}   | {rpe:d} ')

    def display_SPI_map(self) -> None:
        self.logger.log("============================================================")
        self.logger.log("SPI Flash Map")
        self.logger.log("------------------------------------------------------------")
        self.logger.log('')
        self.display_BIOS_region()
        self.logger.log('')
        self.display_SPI_Flash_Regions()
        self.logger.log('')
        self.display_SPI_Flash_Descriptor()
        self.logger.log('')
        self.display_SPI_opcode_info()
        self.logger.log('')
        self.logger.log("============================================================")
        self.logger.log("SPI Flash Protection")
        self.logger.log("------------------------------------------------------------")
        self.logger.log('')
        self.display_SPI_Ranges_Access_Permissions()
        self.logger.log('')
        self.logger.log("BIOS Region Write Protection")
        self.logger.log("------------------------------------------------------------")
        self.display_BIOS_write_protection()
        self.logger.log('')
        self.display_SPI_Protected_Ranges()
        self.logger.log('')

    ##############################################################################################################
    # BIOS Write Protection
    ##############################################################################################################

    def display_BIOS_write_protection(self) -> None:
        if self.cs.register.is_defined('BC'):
            reg_value = self.cs.register.read('BC')
            self.cs.register.print('BC', reg_value)
        else:
            if self.logger.HAL:
                self.logger.log_error("Could not locate the definition of 'BIOS Control' register..")

    def disable_BIOS_write_protection(self) -> bool:
        if self.logger.HAL:
            self.display_BIOS_write_protection()
        ble = self.cs.control.get('BiosLockEnable')
        bioswe = self.cs.control.get('BiosWriteEnable')
        smmbwp = self.cs.control.get('SmmBiosWriteProtection')

        if smmbwp == 1:
            self.logger.log_hal("[spi] SMM BIOS write protection (SmmBiosWriteProtection) is enabled")

        if bioswe == 1:
            self.logger.log_hal("[spi] BIOS write protection (BiosWriteEnable) is not enabled")
            return True
        elif ble == 0:
            self.logger.log_hal("[spi] BIOS write protection is enabled but not locked. Disabling..")
        else:  # bioswe == 0 and ble == 1
            self.logger.log_hal("[spi] BIOS write protection is enabled. Attempting to disable..")

        # Set BiosWriteEnable control bit
        self.cs.control.set('BiosWriteEnable', 1)

        # read BiosWriteEnable back to check if BIOS writes are enabled
        bioswe = self.cs.control.get('BiosWriteEnable')

        if self.logger.HAL:
            self.display_BIOS_write_protection()
        if self.logger.HAL:
            protection = 'disabled' if bioswe else 'still enabled'
            self.logger.log_important(f'BIOS write protection is {protection} (BiosWriteEnable = {bioswe:d})')

        return (bioswe == 1)

    ##############################################################################################################
    # SPI Controller access functions
    ##############################################################################################################

    def _wait_SPI_flash_cycle_done(self) -> bool:
        self.logger.log_hal('[spi] Wait for SPI cycle ready/done...')
        hsfsts = 0
        cycle_done = False

        for i in range(1000):
            # time.sleep(0.001)
            hsfsts = self.spi_reg_read(self.hsfs_off, 1)

            # cycle_done = (hsfsts & Cfg.Cfg.PCH_RCBA_SPI_HSFSTS_FDONE) and (0 == (hsfsts & Cfg.PCH_RCBA_SPI_HSFSTS_SCIP))
            cycle_done = not (hsfsts & PCH_RCBA_SPI_HSFSTS_SCIP)
            if cycle_done:
                break

        if not cycle_done:
            self.logger.log_hal('[spi] SPI cycle still in progress. Waiting 0.1 sec...')
            time.sleep(0.1)
            hsfsts = self.spi_reg_read(self.hsfs_off, 1)
            cycle_done = not (hsfsts & PCH_RCBA_SPI_HSFSTS_SCIP)

        if cycle_done:
            self.logger.log_hal('[spi] Clear FDONE/FCERR/AEL bits...')
            self.spi_reg_write(self.hsfs_off, HSFSTS_CLEAR, 1)
            hsfsts = self.spi_reg_read(self.hsfs_off, 1)
            cycle_done = not ((hsfsts & PCH_RCBA_SPI_HSFSTS_AEL) or (hsfsts & PCH_RCBA_SPI_HSFSTS_FCERR))

        self.logger.log_hal(f'[spi] HSFS: 0x{hsfsts:02X}')

        return cycle_done

    def _send_spi_cycle(self, hsfctl_spi_cycle_cmd: int, dbc: int, spi_fla: int) -> bool:
        self.logger.log_hal(f'[spi] > Send SPI cycle 0x{hsfctl_spi_cycle_cmd:x} to address 0x{spi_fla:08X}')

        # No need to check for SPI cycle DONE status before each cycle
        # DONE status is checked once before entire SPI operation

        self.spi_reg_write(self.faddr_off, (spi_fla & PCH_RCBA_SPI_FADDR_MASK))
        # Other options ;)
        # chipsec.chipset.write_register( self.cs, "FADDR", (spi_fla & Cfg.PCH_RCBA_SPI_FADDR_MASK) )
        # write_MMIO_reg( self.cs, spi_base, self.faddr_off, (spi_fla & Cfg.PCH_RCBA_SPI_FADDR_MASK) )
        # self.cs.mem.write_physical_mem_dword( spi_base + self.faddr_off, (spi_fla & Cfg.PCH_RCBA_SPI_FADDR_MASK) )

        if self.logger.HAL:
            _faddr = self.spi_reg_read(self.faddr_off)
            self.logger.log(f'[spi] FADDR: 0x{_faddr:08X}')

        self.logger.log_hal(f'[spi] SPI cycle GO (DBC <- 0x{dbc:02X}, HSFC <- 0x{hsfctl_spi_cycle_cmd:x})')

        if (HSFCTL_ERASE_CYCLE != hsfctl_spi_cycle_cmd):
            self.spi_reg_write(self.hsfc_off + 0x1, dbc, 1)

        self.spi_reg_write(self.hsfc_off, hsfctl_spi_cycle_cmd, 1)
        # self.spi_reg_write( self.hsfc_off, ((dbc<<8)|hsfctl_spi_cycle_cmd), 2 )

        # Read HSFC back (logging only)
        if self.logger.HAL:
            _hsfc = self.spi_reg_read(self.hsfc_off, 1)
            self.logger.log(f'[spi] HSFC: 0x{_hsfc:04X}')

        cycle_done = self._wait_SPI_flash_cycle_done()
        if not cycle_done:
            self.logger.log_warning("SPI cycle not done")
        else:
            self.logger.log_hal('[spi] < SPI cycle done')

        return cycle_done

    def check_hardware_sequencing(self) -> None:
        # Test if the flash decriptor is valid (and hardware sequencing enabled)
        fdv = self.cs.register.read_field('HSFS', 'FDV')
        if fdv == 0:
            self.logger.log_error("HSFS.FDV is 0, hardware sequencing is disabled")
            raise SpiRuntimeError("Chipset does not support hardware sequencing")

    #
    # SPI Flash operations
    #

    def read_spi_to_file(self, spi_fla: int, data_byte_count: int, filename: str) -> bytes:
        buf = self.read_spi(spi_fla, data_byte_count)
        if buf is None:
            return b''
        if filename is not None:
            write_file(filename, buf)
        else:
            print_buffer_bytes(buf, 16)
        return buf

    def write_spi_from_file(self, spi_fla: int, filename: str) -> bool:
        buf = read_file(filename)
        return self.write_spi(spi_fla, buf)
        # return self.write_spi( spi_fla, struct.unpack('B'*len(buf), buf) )

    def read_spi(self, spi_fla: int, data_byte_count: int) -> bytes:

        self.check_hardware_sequencing()

        buf = bytearray()
        dbc = SPI_READ_WRITE_DEF_DBC
        if (data_byte_count >= SPI_READ_WRITE_MAX_DBC):
            dbc = SPI_READ_WRITE_MAX_DBC

        n = data_byte_count // dbc
        r = data_byte_count % dbc
        if self.logger.UTIL_TRACE or self.logger.HAL:
            self.logger.log(f'[spi] Reading 0x{data_byte_count:x} bytes from SPI at FLA = 0x{spi_fla:x} (in {n:d} 0x{dbc:x}-byte chunks + 0x{r:x}-byte remainder)')

        cycle_done = self._wait_SPI_flash_cycle_done()
        if not cycle_done:
            self.logger.log_error("SPI cycle not ready")
            return b''
        progress_max = DEFAULT_PROGRESS_MAX if n > DEFAULT_PROGRESS_MAX or n <= 0 else n
        progress_tick = n // progress_max
        if self.logger.UTIL_TRACE and not self.logger.HAL:
            self.logger.log(f'_{"_" * progress_max}|')
        for i in range(n):
            if self.logger.UTIL_TRACE and not self.logger.HAL and ((i % progress_tick) == 0):
                self.logger.log_inline('-')
            self.logger.log_hal(f'[spi] Reading chunk {i:d} of 0x{dbc:x} bytes from 0x{spi_fla + i * dbc:x}')
            if not self._send_spi_cycle(HSFCTL_READ_CYCLE, dbc - 1, spi_fla + i * dbc):
                self.logger.log_error("SPI flash read failed")
            else:
                for fdata_idx in range(0, dbc // 4):
                    dword_value = self.spi_reg_read(self.fdata0_off + fdata_idx * 4)
                    if self.logger.HAL:
                        self.logger.log(f'[spi] FDATA00 + 0x{fdata_idx * 4:x}: 0x{dword_value:x}')
                    buf += struct.pack("I", dword_value)
        if self.logger.UTIL_TRACE:
            self.logger.log('\nSPI read complete')

        if (0 != r):
            self.logger.log_hal(f'[spi] Reading remaining 0x{r:x} bytes from 0x{spi_fla + n * dbc:x}')
            if not self._send_spi_cycle(HSFCTL_READ_CYCLE, r - 1, spi_fla + n * dbc):
                self.logger.log_error("SPI flash read failed")
            else:
                t = 4
                n_dwords = (r + 3) // 4
                for fdata_idx in range(0, n_dwords):
                    dword_value = self.spi_reg_read(self.fdata0_off + fdata_idx * 4)
                    if self.logger.HAL:
                        self.logger.log(f'[spi] FDATA00 + 0x{fdata_idx * 4:x}: 0x{dword_value:08X}')
                    if (fdata_idx == (n_dwords - 1)) and (0 != r % 4):
                        t = r % 4
                    for j in range(t):
                        buf += struct.pack('B', (dword_value >> (8 * j)) & 0xff)

        self.logger.log_hal('[spi] Buffer read from SPI:')
        if self.logger.HAL:
            print_buffer_bytes(buf)

        return bytes(buf)

    def write_spi(self, spi_fla: int, buf: bytes) -> bool:

        self.check_hardware_sequencing()

        write_ok = True
        data_byte_count = len(buf)
        dbc = 4
        n = data_byte_count // dbc
        r = data_byte_count % dbc
        if self.logger.UTIL_TRACE or self.logger.HAL:
            self.logger.log(f'[spi] Writing 0x{data_byte_count:x} bytes to SPI at FLA = 0x{spi_fla:x} (in {n:d} 0x{dbc:x}-byte chunks + 0x{r:x}-byte remainder)')

        cycle_done = self._wait_SPI_flash_cycle_done()
        if not cycle_done:
            self.logger.log_error("SPI cycle not ready")
            return False

        for i in range(n):
            if self.logger.UTIL_TRACE or self.logger.HAL:
                self.logger.log(f'[spi] Writing chunk {i:d} of 0x{dbc:x} bytes to 0x{spi_fla + i * dbc:x}')
            dword_value = ((buf[i * dbc + 3]) << 24) | ((buf[i * dbc + 2]) << 16) | ((buf[i * dbc + 1]) << 8) | (buf[i * dbc])
            if self.logger.HAL:
                self.logger.log(f'[spi] in FDATA00 = 0x{dword_value:08X}')
            self.spi_reg_write(self.fdata0_off, dword_value)
            if not self._send_spi_cycle(HSFCTL_WRITE_CYCLE, dbc - 1, spi_fla + i * dbc):
                write_ok = False
                self.logger.log_error("SPI flash write cycle failed")

        if (0 != r):
            if self.logger.UTIL_TRACE or self.logger.HAL:
                self.logger.log(f'[spi] Writing remaining 0x{r:x} bytes to FLA = 0x{spi_fla + n * dbc:x}')
            dword_value = 0
            for j in range(r):
                dword_value |= (buf[n * dbc + j] << 8 * j)
            if self.logger.HAL:
                self.logger.log(f'[spi] in FDATA00 = 0x{dword_value:08X}')
            self.spi_reg_write(self.fdata0_off, dword_value)
            if not self._send_spi_cycle(HSFCTL_WRITE_CYCLE, r - 1, spi_fla + n * dbc):
                write_ok = False
                self.logger.log_error("SPI flash write cycle failed")

        return write_ok

    def erase_spi_block(self, spi_fla: int) -> bool:

        self.check_hardware_sequencing()

        if self.logger.UTIL_TRACE or self.logger.HAL:
            self.logger.log(f'[spi] Erasing SPI Flash block @ 0x{spi_fla:x}')

        cycle_done = self._wait_SPI_flash_cycle_done()
        if not cycle_done:
            self.logger.log_error("SPI cycle not ready")
            return cycle_done

        erase_ok = self._send_spi_cycle(HSFCTL_ERASE_CYCLE, 0, spi_fla)
        if not erase_ok:
            self.logger.log_error("SPI Flash erase cycle failed")

        return erase_ok

    #
    # SPI SFDP operations
    #
    def ptmesg(self, offset: int) -> int:
        self.spi_reg_write(self.bios_ptinx, offset)
        self.spi_reg_read(self.bios_ptinx)
        return self.spi_reg_read(self.bios_ptdata)

    def get_SPI_SFDP(self) -> bool:
        ret = False
        for component in range(0, 2):
            self.logger.log(f'Scanning for Flash device {component + 1:d}')
            offset = 0x0000 | (component << 14)
            sfdp_signature = self.ptmesg(offset)
            if sfdp_signature == SFDP_HEADER:
                self.logger.log(f'  * Found valid SFDP header for Flash device {component + 1:d}')
                ret = True
            else:
                self.logger.log(f"  * Didn't find a valid SFDP header for Flash device {component + 1:d}")
                continue
            # Increment offset to read second dword of SFDP header structure
            sfdp_data = self.ptmesg(offset + 0x4)
            sfdp_minor_version = sfdp_data & 0xFF
            sfdp_major_version = (sfdp_data >> 8) & 0xFF
            self.logger.log(f'    SFDP version number: {sfdp_major_version}.{sfdp_minor_version}')
            num_of_param_headers = ((sfdp_data >> 16) & 0xFF) + 1
            self.logger.log(f'    Number of parameter headers: {num_of_param_headers:d}')
            # Set offset to read 1st Parameter Table in the SFDP header structure
            offset = offset | 0x1000
            parameter_1 = self.ptmesg(offset)
            param1_minor_version = (parameter_1 >> 8) & 0xFF
            param1_major_version = (parameter_1 >> 16) & 0xFF
            param1_length = (parameter_1 >> 24) & 0xFF
            self.logger.log("  * Parameter Header 1 (JEDEC)")
            self.logger.log(f'    ** Parameter version number: {param1_major_version}.{param1_minor_version}')
            self.logger.log(f'    ** Parameter length in double words: {hex(param1_length)}')
            if (num_of_param_headers > 1) and self.cs.register.has_field('HSFS', 'FCYCLE'):
                self.check_hardware_sequencing()
                self.spi_reg_write(self.fdata12_off, 0x00000000)
                self.spi_reg_write(self.fdata13_off, 0x00000000)
                self.spi_reg_write(self.fdata14_off, 0x00000000)
                self.spi_reg_write(self.fdata15_off, 0x00000000)
                if not self._send_spi_cycle(HSFCTL_SFDP_CYCLE, 0x3F, 0):
                    self.logger.log_error('SPI SFDP signature cycle failed')
                    continue
                pTable_offset_list = []
                pTable_length = []
                # Calculate which fdata_offset registers to read, based on number of parameter headers present
                for i in range(1, num_of_param_headers):
                    self.logger.log(f'  * Parameter Header:{i + 1:d}')
                    data_reg_1 = f'self.fdata{str(2 + (2 * i))}_off'
                    data_reg_2 = f'self.fdata{str(2 + (2 * i) + 1)}_off'
                    data_dword_1 = self.spi_reg_read(eval(data_reg_1))
                    data_dword_2 = self.spi_reg_read(eval(data_reg_2))
                    id_manuf = (data_dword_2 & 0xFF000000) >> 16 | (data_dword_1 & 0xFF)
                    param_minor_version = (data_dword_1 >> 8) & 0xFF
                    param_major_version = (data_dword_1 >> 16) & 0xFF
                    param_length = (data_dword_1 >> 24) & 0xFF
                    param_table_pointer = (data_dword_2 & 0x00FFFFFF)
                    self.logger.log(f'    ** Parameter version number:{param_major_version}.{param_minor_version}')
                    self.logger.log(f'    ** Parameter length in double words: {hex(param_length)}')
                    self.logger.log(f'    ** Parameter ID: {hex(id_manuf)}')
                    self.logger.log(f'    ** Parameter Table Pointer(byte address): {hex(param_table_pointer)} ')
                    pTable_offset_list.append(param_table_pointer)
                    pTable_length.append(param_length)
            offset = 0x0000 | (component << 14)
            # Set offset to read 1st Parameter table ( JEDEC Basic Flash Parameter Table) content and Parse it
            offset = offset | 0x2000
            self.logger.log("                                ")
            self.logger.log("  * 1'st Parameter Table Content ")
            for count in range(1, param1_length + 1):
                sfdp_data = self.ptmesg(offset)
                offset += 4
                self.cs.register.print(f'DWORD{count}', sfdp_data)
        return ret

    #
    # SPI JEDEC ID operations
    #

    def get_SPI_JEDEC_ID(self) -> int:

        if self.cs.register.has_field('HSFS', 'FCYCLE'):
            self.check_hardware_sequencing()

            if not self._send_spi_cycle(HSFCTL_JEDEC_CYCLE, 4, 0):
                self.logger.log_error('SPI JEDEC ID cycle failed')
            id = self.spi_reg_read(self.fdata0_off)
        else:
            return False

        return ((id & 0xFF) << 16) | (id & 0xFF00) | ((id >> 16) & 0xFF)

    def get_SPI_JEDEC_ID_decoded(self) -> Tuple[int, str, str]:

        jedec_id = self.get_SPI_JEDEC_ID()
        if jedec_id is False:
            return (False, '', '')
        manu = JEDEC_ID.MANUFACTURER.get((jedec_id >> 16) & 0xff, 'Unknown')
        part = JEDEC_ID.DEVICE.get(jedec_id, 'Unknown')

        return (jedec_id, manu, part)



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

# -------------------------------------------------------------------------------
#
# CHIPSEC: Platform Hardware Security Assessment Framework
#
# -------------------------------------------------------------------------------

"""
Working with Intel processor Integrated Graphics Device (IGD)

usage:
    >>> gfx_aperture_dma_read(0x80000000, 0x100)
"""

from typing import Optional, Tuple
from chipsec.hal import hal_base
from chipsec.library.logger import print_buffer_bytes


class IGD(hal_base.HALBase):

    def __init__(self, cs):
        super(IGD, self).__init__(cs)
        self.helper = cs.helper
        self.is_legacy = False
        self.enabled = None

    def __identify_device(self) -> Tuple[bool, bool]:
        if self.enabled is None:
            try:
                self.dev_id = self.cs.register.read("PCI0.2.0_DID")
                self.enabled = (self.dev_id != 0xFFFF)
                if self.enabled:
                    self.is_legacy = bool(self.dev_id < 0x1600)
            except Exception:
                self.enabled = False

        return (self.enabled, self.is_legacy)

    def is_enabled(self) -> bool:
        if self.cs.register.has_field("PCI0.0.0_DEVEN", "D2EN") and self.cs.register.has_field("PCI0.0.0_CAPID0_A", "IGD"):
            if self.cs.register.read_field("PCI0.0.0_DEVEN", "D2EN") == 1 and self.cs.register.read_field("PCI0.0.0_CAPID0_A", "IGD") == 0:
                return True
        elif self.cs.register.has_field("PCI0.0.0_DEVEN", "D2EN"):
            if self.cs.register.read_field("PCI0.0.0_DEVEN", "D2EN") == 1:
                return True
        elif self.cs.register.has_field("PCI0.0.0_CAPID0_A", "IGD"):
            if self.cs.register.read_field("PCI0.0.0_CAPID0_A", "IGD") == 0:
                return True
        return self.is_device_enabled()

    def is_device_enabled(self) -> bool:
        enabled, _ = self.__identify_device()
        return enabled

    def is_legacy_gen(self) -> bool:
        _, legacy = self.__identify_device()
        return legacy

    def get_GMADR(self) -> int:
        base, _ = self.cs.mmio.get_MMIO_BAR_base_address('GMADR')
        self.logger.log_hal(f'[igd] Aperture (GMADR): 0x{base:016X}')
        return base

    def get_GTTMMADR(self) -> int:
        base, _ = self.cs.mmio.get_MMIO_BAR_base_address('GTTMMADR')
        self.logger.log_hal(f'[igd] Graphics MMIO and GTT (GTTMMADR): 0x{base:016X}')
        return base

    def get_GGTT_base(self) -> int:
        gtt_off = 0x200000 if self.is_legacy_gen() else 0x800000
        return self.get_GTTMMADR() + gtt_off

    def get_PTE_size(self) -> int:
        return 4 if self.is_legacy_gen() else 8

    def read_GGTT_PTE(self, pte_num: int) -> int:
        gtt_base = self.get_GGTT_base()
        reg_off = (self.get_PTE_size() * pte_num)

        pte_lo = self.cs.mmio.read_MMIO_reg(gtt_base, reg_off)
        pte_hi = 0
        if self.get_PTE_size() == 8:
            pte_hi = self.cs.mmio.read_MMIO_reg(gtt_base, reg_off + 4)
        return (pte_lo | (pte_hi << 32))

    def write_GGTT_PTE(self, pte_num: int, pte: int) -> int:
        gtt_base = self.get_GGTT_base()
        self.cs.mmio.write_MMIO_reg(gtt_base, self.get_PTE_size() * pte_num, pte & 0xFFFFFFFF)
        if self.get_PTE_size() == 8:
            self.cs.mmio.write_MMIO_reg(gtt_base, self.get_PTE_size() * pte_num + 4, pte >> 32)
        return pte

    def write_GGTT_PTE_from_PA(self, pte_num: int, pa: int) -> int:
        pte = self.get_GGTT_PTE_from_PA(pa)
        gtt_base = self.get_GGTT_base()
        self.cs.mmio.write_MMIO_reg(gtt_base, self.get_PTE_size() * pte_num, pte & 0xFFFFFFFF)
        if self.get_PTE_size() == 8:
            self.cs.mmio.write_MMIO_reg(gtt_base, self.get_PTE_size() * pte_num + 4, pte >> 32)
        return pte

    def dump_GGTT_PTEs(self, num: int) -> None:
        gtt_base = self.get_GGTT_base()
        self.logger.log('[igd] Global GTT contents:')
        ptes = self.cs.mmio.read_MMIO(gtt_base, num * self.get_PTE_size())
        pte_num = 0
        for pte in ptes:
            self.logger.log(f'PTE[{pte_num:03d}]: {pte:08X}')
            pte_num = pte_num + 1

    def get_GGTT_PTE_from_PA(self, pa: int) -> int:
        if self.is_legacy_gen():
            return self.get_GGTT_PTE_from_PA_legacy(pa)
        else:
            return self.get_GGTT_PTE_from_PA_gen8(pa)

    def get_GGTT_PTE_from_PA_legacy(self, pa: int) -> int:
        #
        # GTT PTE format:
        # 0     - valid
        # 2:1   - cache type (00 - reserved, 01 - UC, 10 - LLC only, 11 - MLC/LLC)
        # 3     - GFDT
        # 11:4  - PA bits 39:32
        # 31:12 - PA bits 31:12
        #
        return ((pa & 0xFFFFF000) | ((pa >> 32 & 0xFF) << 4) | 0x3)

    def get_PA_from_PTE_legacy(self, pte: int) -> int:
        return (((pte & 0x00000FF0) << 28) | (pte & 0xFFFFF000))

    def get_GGTT_PTE_from_PA_gen8(self, pa: int) -> int:
        return ((pa & ~0xFFF) | 0x1)

    def get_PA_from_PTE_gen8(self, pte: int) -> int:
        return (pte & ~0xFFF)

    def get_PA_from_PTE(self, pte: int) -> int:
        if self.is_legacy_gen():
            return self.get_PA_from_PTE_legacy(pte)
        else:
            return self.get_PA_from_PTE_gen8(pte)

    def gfx_aperture_dma_read_write(self, address: int, size: int = 0x4, value: Optional[bytes] = None, pte_num: int = 0) -> bytes:
        r = 0
        pages = 0

        gmadr = self.get_GMADR()
        off = address % 0x1000
        h = 0x1000 - off
        igd_addr = gmadr + pte_num * 0x1000
        pte_orig = self.read_GGTT_PTE(pte_num)

        self.logger.log_hal(f'[igd] Reading 0x{size:X} bytes at PA 0x{address:016X} through IGD aperture (DMA) using PTE{pte_num:d}')
        self.logger.log_hal(f'[igd] GFx aperture (GMADR): 0x{gmadr:016X}')
        self.logger.log_hal(f'[igd] GFx GTT base        : 0x{self.get_GGTT_base():016X}')
        self.logger.log_hal(f'[igd] Original GTT PTE{pte_num:03d}: 0x{pte_orig:08X}')

        if (h > 0) and (size > h):
            r = (size - h) % 0x1000
            pages = 2 + (size - h) // 0x1000
        else:
            r = size % 0x1000
            pages = 1 + size // 0x1000

        N = pages
        self.logger.log_hal(f'[igd] Pages = 0x{pages:X}, r = 0x{r:X}, N = {N:d}')

        self.logger.log_hal(f'[igd] Original data at address 0x{address:016X}:')
        if self.logger.HAL:
            print_buffer_bytes(self.cs.mem.read_physical_mem(address, size))

        buffer = b''
        pa = address
        for p in range(N):
            pte = self.get_GGTT_PTE_from_PA(pa)
            if self.logger.HAL:
                self.logger.log(f'[igd] GFx PTE for address 0x{address:016X}: 0x{pte:08X}')
            self.write_GGTT_PTE(pte_num, pte)
            if (p == 0):
                pa_off = off
                size = h if (pa_off > 0) else 0x1000
            else:
                pa_off = 0
            if (p == N - 1):
                size = r if (r > 0) else 0x1000
            if value is None:
                self.logger.log_hal(f'[igd] Reading 0x{size:X} bytes at 0x{pa:016X} through GFx aperture 0x{igd_addr + pa_off:016X}...')
                page = self.cs.mem.read_physical_mem(igd_addr + pa_off, size)
                buffer += page
                if self.logger.HAL:
                    print_buffer_bytes(page[:size])
            else:
                self.logger.log_hal(f'[igd] Writing 0x{size:X} bytes to 0x{pa:016X} through GFx aperture 0x{igd_addr + pa_off:016X}...')
                page = value[p * 0x1000:p * 0x1000 + size]
                self.cs.mem.write_physical_mem(igd_addr + pa_off, size, page)
                if self.logger.HAL:
                    print_buffer_bytes(page)
            pa += size

        # restore original PTE
        self.logger.log_hal(f'[igd] Restoring GFx PTE{pte_num:d} 0x{pte_orig:X}...')
        self.write_GGTT_PTE(pte_num, pte_orig)

        return buffer



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


"""
Hyper-V VMBus synthetic device generic fuzzer

Usage:

  Print channel offers:

  ``chipsec_main.py -i -m tools.vmm.hv.synth_dev -a info``

  Fuzzing device with specified relid:

  ``chipsec_main.py -i -m tools.vmm.hv.synth_dev -a fuzz,<relid> -l log.txt``

Note: the fuzzer is incompatible with native VMBus driver (``vmbus.sys``). To use it, remove ``vmbus.sys``
"""

import sys
import traceback
from struct import pack
from chipsec.library.returncode import ModuleResult
from chipsec.module_common import BaseModule, CPU
from chipsec.modules.tools.vmm.common import session_logger, get_int_arg
from chipsec.modules.tools.vmm.hv.define import VMBUS_DATA_PACKET_FLAG_COMPLETION_REQUESTED, vm_pkt
from chipsec.modules.tools.vmm.hv.vmbus import RingBuffer, VMBusDiscovery

sys.stdout = session_logger(True, 'synth_dev')

TAGS = [CPU]
METADATA_TAGS = ['OPENSOURCE', 'IA', 'TOOLS', 'VMM', 'HV', 'SYNTH_DEV']


class VMBusDeviceFuzzer(VMBusDiscovery):
    def __init__(self):
        VMBusDiscovery.__init__(self)
        self.responses = {}

    def send_1(self, relid, messages, info, order):
        if len(messages) > 0:
            msg_sent = messages.pop(0)
            vmpkt_datainband = list(vm_pkt.keys())[list(vm_pkt.values()).index('VM_PKT_DATA_INBAND')]
            self.vmbus_sendpacket(relid, msg_sent, 0x0, vmpkt_datainband, VMBUS_DATA_PACKET_FLAG_COMPLETION_REQUESTED)
            msg_recv = self.vmbus_recvpacket(relid)
            if msg_recv != '':
                (msg1, msg2) = (msg_recv, msg_sent) if order else (msg_sent, msg_recv)
                if msg1 not in info:
                    info[msg1] = {'next': {}, 'count': 0, 'message': ''}
                info[msg1]['count'] += 1
                info[msg1]['message'] = msg2
                info[msg1]['next'] = self.send_1(relid, messages, info[msg1]['next'], order)
        return info

    def device_fuzzing(self, relid):
        for x in range(1, 0x100):
            for a in range(0, 0x100):
                self.ringbuffers[relid].ringbuffer_init()
                self.vmbus_establish_gpadl(relid, self.ringbuffers[relid].gpadl, self.ringbuffers[relid].pfn)
                self.vmbus_open(relid, self.ringbuffers[relid].gpadl, self.ringbuffers[relid].send_size)
                msg = pack('<LL', x, ((a & 0xf0) << 12) | (a % 0x0f))
                try:
                    self.responses = self.send_1(relid, [msg], self.responses, True)
                finally:
                    self.vmbus_close(relid)
                    self.vmbus_teardown_gpadl(relid, self.ringbuffers[relid].gpadl)
        return

    def print_1(self, info, indent=0):
        if len(info) == 0:
            return
        for i in self.responses:
            self.msg(f'{"  " * indent}{i.hex():20}:{info[i]["message"].hex():20}  {info[i]["count"]:4d}')
            self.print_1(info[i]['next'], indent + 1)
        return

    def print_statistics(self):
        self.msg('Response statistics:')
        self.print_1(self.responses)
        return


class synth_dev(BaseModule):
    def __init__(self):
        BaseModule.__init__(self)

    def usage(self):
        self.logger.log(__doc__)
        return

    def run(self, module_argv):
        self.logger.start_test("Hyper-V VMBus synthetic device fuzzer")

        command = module_argv[0] if len(module_argv) > 0 and module_argv[0] != '' else 'none'
        relid = get_int_arg(module_argv[1]) if len(module_argv) > 1 and module_argv[1] != '' else 0x5

        vb = VMBusDeviceFuzzer()
        vb.debug = False
        vb.vmbus_init()
        try:
            vb.vmbus_connect()
            vb.vmbus_request_offers()

            if relid not in [value['child_relid'] for value in vb.offer_channels.values()]:
                vb.fatal(f'child relid #{relid:d} has not been found!')

            vb.ringbuffers[relid] = RingBuffer()
            vb.ringbuffers[relid].debug = False
            vb.ringbuffers[relid].ringbuffer_alloc(4)
            vb.ringbuffers[relid].gpadl = vb.vmbus_get_next_gpadl()

            if command == 'info':
                vb.vmbus_establish_gpadl(relid, vb.ringbuffers[relid].gpadl, vb.ringbuffers[relid].pfn)
                vb.vmbus_open(relid, vb.ringbuffers[relid].gpadl, vb.ringbuffers[relid].send_size)
                vb.print_offer_channels()
                vb.print_created_gpadl()
                vb.print_open_channels()
                vb.vmbus_close(relid)
                vb.vmbus_teardown_gpadl(relid, vb.ringbuffers[relid].gpadl)
            elif command == 'fuzz':
                vb.promt = f'DEVICE {relid:02d}'
                vb.msg('Fuzzing VMBus devices ...')
                vb.device_fuzzing(relid)
                vb.print_statistics()
            else:
                self.usage()

        except KeyboardInterrupt:
            self.logger.log('***** Control-C *****')
        except Exception:
            self.logger.log('\n\n')
            traceback.print_exc()
            self.logger.log('\n\n')
        finally:
            vb.vmbus_rescind_all_offers()
            del vb
        self.result.setStatusBit(self.result.status.SUCCESS)
        return self.result.getReturnCode(ModuleResult.PASSED)



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


"""
CPU related functionality

"""
from typing import Dict, List, Tuple, Optional
from chipsec.hal import acpi, hal_base, paging
from chipsec.library.logger import logger

VMM_NONE = 0
VMM_XEN = 0x1
VMM_HYPER_V = 0x2
VMM_VMWARE = 0x3
VMM_KVM = 0x4


########################################################################################################
#
# CORES HAL Component
#
########################################################################################################

class CPU(hal_base.HALBase):
    def __init__(self, cs):
        super(CPU, self).__init__(cs)
        self.helper = cs.helper

    def read_cr(self, cpu_thread_id: int, cr_number: int) -> int:
        value = self.helper.read_cr(cpu_thread_id, cr_number)
        logger().log_hal(f'[cpu{cpu_thread_id:d}] read CR{cr_number:d}: value = 0x{value:08X}')
        return value

    def write_cr(self, cpu_thread_id: int, cr_number: int, value: int) -> int:
        logger().log_hal(f'[cpu{cpu_thread_id:d}] write CR{cr_number:d}: value = 0x{value:08X}')
        status = self.helper.write_cr(cpu_thread_id, cr_number, value)
        return status

    def cpuid(self, eax: int, ecx: int) -> Tuple[int, int, int, int]:
        logger().log_hal(f'[cpu] CPUID in : EAX=0x{eax:08X}, ECX=0x{ecx:08X}')
        (eax, ebx, ecx, edx) = self.helper.cpuid(eax, ecx)
        logger().log_hal(f'[cpu] CPUID out: EAX=0x{eax:08X}, EBX=0x{ebx:08X}, ECX=0x{ecx:08X}, EDX=0x{edx:08X}')
        return (eax, ebx, ecx, edx)

    # Using cpuid check if running under vmm control
    def check_vmm(self) -> int:
        # check Hypervisor Present
        (_, ebx, ecx, edx) = self.cpuid(0x01, 0)
        if (ecx & 0x80000000):
            (_, ebx, ecx, edx) = self.cpuid(0x40000000, 0)
            is_xen = ((ebx == 0x566e6558) and (ecx == 0x65584d4d) and (edx == 0x4d4d566e))
            if is_xen:
                return VMM_XEN
            is_hyperv = ((ebx == 0x7263694D) and (ecx == 0x666F736F) and (edx == 0x76482074))
            if is_hyperv:
                return VMM_HYPER_V
            is_vmware = ((ebx == 0x61774d56) and (ecx == 0x4d566572) and (edx == 0x65726177))
            if is_vmware:
                return VMM_VMWARE
            is_kvm = ((ebx == 0x4b4d564b) and (ecx == 0x564b4d56) and (edx == 0x0000004d))
            if is_kvm:
                return VMM_KVM
        return VMM_NONE

    # Using CPUID we can determine if Hyper-Threading is enabled in the CPU
    def is_HT_active(self) -> bool:
        logical_processor_per_core = self.get_number_logical_processor_per_core()
        return logical_processor_per_core > 1

    # Using the CPUID we determine the number of logical processors per core
    def get_number_logical_processor_per_core(self) -> int:
        (_, ebx, _, _) = self.cpuid(0x0b, 0x0)
        return ebx

    # Using CPUID we can determine the number of logical processors per package
    def get_number_logical_processor_per_package(self) -> int:
        (_, ebx, _, _) = self.cpuid(0x0b, 0x1)
        return ebx

    # Using CPUID we can determine the number of physical processors per package
    def get_number_physical_processor_per_package(self) -> int:
        logical_processor_per_core = self.get_number_logical_processor_per_core()
        logical_processor_per_package = self.get_number_logical_processor_per_package()
        return (logical_processor_per_package // logical_processor_per_core)

    # determine number of logical processors in the core
    def get_number_threads_from_APIC_table(self) -> int:
        _acpi = acpi.ACPI(self.cs)
        dACPIID = {}
        for apic in _acpi.get_parse_ACPI_table(acpi.ACPI_TABLE_SIG_APIC):  # (table_header, APIC_object, table_header_blob, table_blob)
            _, APIC_object, _, _ = apic
            for structure in APIC_object.apic_structs:
                if 0x00 == structure.Type:
                    if not structure.ACICID in dACPIID:
                        if 1 == structure.Flags:
                            dACPIID[structure.APICID] = structure.ACPIProcID
        return len(dACPIID)

    # determine the cpu threads location within a package/core
    def get_cpu_topology(self) -> Dict[str, Dict[int, List[int]]]:
        num_threads = self.cs.helper.get_threads_count()
        packages: Dict[int, List[int]] = {}
        cores: Dict[int, List[int]] = {}
        for thread in range(num_threads):
            if num_threads > 1:
                self.logger.log_hal(f'Setting affinity to: {thread:d}')
                self.cs.helper.set_affinity(thread)
            eax = 0xb   # cpuid leaf 0B contains x2apic info
            ecx = 1     # ecx 1 will get us pkg_id in edx after shifting right by _eax
            (_eax, _, _, _edx) = self.cs.cpu.cpuid(eax, ecx)
            pkg_id = _edx >> (_eax & 0xf)
            if pkg_id not in packages:
                packages[pkg_id] = []
            packages[pkg_id].append(thread)

            ecx = 0     # ecx 0 will get us the core_id in edx after shifting right by _eax
            (_eax, _, _, _edx) = self.cs.cpu.cpuid(eax, ecx)
            core_id = _edx >> (_eax & 0xf)
            if core_id not in cores:
                cores[core_id] = []
            cores[core_id].append(thread)
            self.logger.log_hal(f'pkg id is {pkg_id:x}')
            self.logger.log_hal(f'core id is {core_id:x}')
        topology = {'packages': packages, 'cores': cores}
        return topology

    # determine number of physical sockets using the CPUID and APIC ACPI table
    def get_number_sockets_from_APIC_table(self) -> int:
        number_threads = self.get_number_threads_from_APIC_table()
        logical_processor_per_package = self.get_number_logical_processor_per_package()
        return (number_threads // logical_processor_per_package)

    #
    # Return SMRR MSR physical base and mask
    #
    def get_SMRR(self) -> Tuple[int, int]:
        if self.cs.is_intel():
            smrambase = self.cs.register.read_field('IA32_SMRR_PHYSBASE', 'PhysBase', True)
            smrrmask = self.cs.register.read_field('IA32_SMRR_PHYSMASK', 'PhysMask', True)
        elif self.cs.is_amd():
            smrambase = self.cs.register.read_field('SMM_BASE', 'SMMBASE', True)
            smrrmask = self.cs.register.read_field('SMMMASK', 'TSEGMASK', True)
        return (smrambase, smrrmask)

    #
    # Return SMRAM region base, limit and size as defined by SMRR
    #
    def get_SMRR_SMRAM(self) -> Tuple[int, int, int]:
        (smram_base, smrrmask) = self.get_SMRR()
        smram_base &= smrrmask
        smram_size = ((~smrrmask) & 0xFFFFFFFF) + 1
        smram_limit = smram_base + smram_size - 1
        return (smram_base, smram_limit, smram_size)

    #
    # Returns TSEG base, limit and size
    #
    def get_TSEG(self) -> Tuple[int, int, int]:
        if self.cs.is_intel():
            if self.cs.is_server():
                # tseg register has base and limit
                tseg_base = self.cs.register.read_field('TSEG_BASE', 'base', preserve_field_position=True)
                tseg_limit = self.cs.register.read_field('TSEG_LIMIT', 'limit', preserve_field_position=True)
                tseg_limit += 0xFFFFF
            else:
                # TSEG base is in TSEGMB, TSEG limit is BGSM - 1
                tseg_base = self.cs.register.read_field('PCI0.0.0_TSEGMB', 'TSEGMB', preserve_field_position=True)
                bgsm = self.cs.register.read_field('PCI0.0.0_BGSM', 'BGSM', preserve_field_position=True)
                tseg_limit = bgsm - 1

            tseg_size = tseg_limit - tseg_base + 1
        elif self.cs.is_amd():
            tseg_base = self.cs.register.read_field('SMMADDR', 'TSEGBASE', preserve_field_position=True)
            tseg_mask = self.cs.register.read_field('SMMMASK', 'TSEGMASK', preserve_field_position=True)
            tseg_size = ((~tseg_mask + 0xFFFFFFFFFFFF) + 1)
            tseg_limit = tseg_base + tseg_size
        return (tseg_base, tseg_limit, tseg_size)

    #
    # Returns SMRAM base from either SMRR MSR or TSEG PCIe config register
    #
    def get_SMRAM(self) -> Tuple[int, int, int]:
        smram_base = None
        smram_limit = None
        smram_size = 0
        try:
            if self.check_SMRR_supported():
                (smram_base, smram_limit, smram_size) = self.get_SMRR_SMRAM()
        except Exception:
            self.logger.log_hal('[cpu] Error using get_SMRR_SMRAM() to get SMRAM!')

        if (smram_base is None) or (smram_limit is None):
            try:
                (smram_base, smram_limit, smram_size) = self.get_TSEG()
            except Exception:
                self.logger.log_hal('[cpu] Error using get_TSEG() to get SMRAM!')
                smram_base = 0
                smram_limit = 0
        return (smram_base, smram_limit, smram_size)

    #
    # Check that SMRR is supported by CPU in IA32_MTRRCAP_MSR[SMRR]
    #
    def check_SMRR_supported(self) -> bool:
        if self.cs.is_intel():
            mtrrcap_msr_reg = self.cs.register.read('MTRRCAP')
            if logger().HAL:
                self.cs.register.print('MTRRCAP', mtrrcap_msr_reg)
            smrr = self.cs.register.get_field('MTRRCAP', mtrrcap_msr_reg, 'SMRR')
            return (1 == smrr)
        elif self.cs.is_amd():
            smmmask_msr_reg = self.cs.register.read('SMMMASK')
            if logger().HAL:
                self.cs.register.print('SMMMASK', smmmask_msr_reg)
            avalid = self.cs.register.get_field('SMMMASK', smmmask_msr_reg, 'AVALID')
            tvalid = self.cs.register.get_field('SMMMASK', smmmask_msr_reg, 'TVALID')
            return (1 == avalid and 1 == tvalid)

    #
    # Dump CPU page tables at specified physical base of paging-directory hierarchy (CR3)
    #
    def dump_page_tables(self, cr3: int, pt_fname: Optional[str] = None) -> None:
        _orig_logname = logger().LOG_FILE_NAME
        hpt = paging.c_ia32e_page_tables(self.cs)
        logger().log_hal(f'[cpu] dumping paging hierarchy at physical base (CR3) = 0x{cr3:08X}...')
        if pt_fname is None:
            pt_fname = f'pt_{cr3:08X}'
        logger().set_log_file(pt_fname)
        hpt.read_pt_and_show_status(pt_fname, 'PT', cr3)
        logger().set_log_file(_orig_logname)
        if hpt.failure:
            logger().log_error('could not dump page tables')

    def dump_page_tables_all(self) -> None:
        for tid in range(self.cs.msr.get_cpu_thread_count()):
            cr3 = self.read_cr(tid, 3)
            logger().log_hal(f'[cpu{tid:d}] found paging hierarchy base (CR3): 0x{cr3:08X}')
            self.dump_page_tables(cr3)



# CHIPSEC: Platform Security Assessment Framework
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
import os
import tempfile
import unittest

from tests.software import mock_helper, util


class TestSPIChipsecUtil(util.TestChipsecUtil):
    """Test the SPI commands exposed by chipsec_utils."""

    def test_spi_info(self):
        """Test to verify the ouput of 'spi info'.

        Validates that BC and FRAP are correctly read.
        """

        self._chipsec_util("spi info", mock_helper.SPIHelper)
        self._assertLogValue("BC", "0xDEADBEEF")
        self._assertLogValue("FRAP", "0xEEEEEEEE")

    def test_spi_dump(self):
        """Test to verify the ouput of 'spi dump'.

        Validates that the flash size is correctly calculated (based on the
        assumption that the BIOS region is last) and match it with the output
        file size.
        """

        fileno, rom_file = tempfile.mkstemp()
        os.close(fileno)
        self._chipsec_util("spi dump {}".format(rom_file), mock_helper.SPIHelper)
        self.assertEqual(os.stat(rom_file).st_size, 0x3000)
        os.remove(rom_file)


if __name__ == '__main__':
    unittest.main()



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2016, Google
# Copyright (c) 2019-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#


from enum import Enum
import traceback

from chipsec.library.logger import logger
from chipsec.library.defines import CHIPSET_CODE_UNKNOWN
from chipsec.testcase import ExitCode


class BaseCommand:

    def __init__(self, argv, cs=None):
        self.argv = argv
        self.logger = logger()
        self.cs = cs
        self.ExitCode = ExitCode.OK

    def run(self) -> None:
        try:
            self.func()
        except Exception:
            self.logger.log_error('An error occured during the execution of the command!')
            self.logger.log_error('Please run with the debug option for further details')
            if logger().DEBUG:
                traceback.print_exc()

    def prerun(self) -> bool:
        if self.requirements().load_config() and self.cs.Cfg.code == CHIPSET_CODE_UNKNOWN:
            self.logger.log_error('This module requires a configuration to be loaded. Please use the "-p XXX" flag to specify a platform configuration to load.')
            return False
        return True

    def set_up(self) -> None:
        pass

    def tear_down(self) -> None:
        pass

    def parse_arguments(self) -> None:
        raise NotImplementedError('sub class should overwrite the parse_arguments() method')

    def requirements(self) -> 'toLoad':
        raise NotImplementedError('sub class should overwrite the requirements() method')


class toLoad(Enum):
    Nil = 0
    Config = 1
    Driver = 2
    All = 3

    def load_config(self) -> bool:
        if self in [self.Config, self.All]:
            return True
        return False

    def load_driver(self) -> bool:
        if self in [self.Driver, self.All]:
            return True
        return False



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2022, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

"""
Checks for RTC memory locks.
Since we do not know what RTC memory will be used for on a specific platform, we return WARNING (rather than FAILED) if the memory is not locked.

Usage:
    ``chipsec_main -m common.rtclock [-a modify]``

    - ``-a modify``: Attempt to modify CMOS values

Examples:
    >>> chipsec_main.py -m common.rtclock
    >>> chipsec_main.py -m common.rtclock -a modify

Registers used:
    - RC.LL
    - RC.UL

.. NOTE::
    - This module will only run on Core platforms
"""

from chipsec.module_common import BaseModule, BIOS, HWCONFIG
from chipsec.library.returncode import ModuleResult
from chipsec.hal.cmos import CMOS
from chipsec.config import CHIPSET_CODE_AVN
from typing import List

TAGS = [BIOS, HWCONFIG]
METADATA_TAGS = ['OPENSOURCE', 'IA', 'COMMON', 'RTCLOCK']


class rtclock(BaseModule):

    def __init__(self):
        BaseModule.__init__(self)
        self.cmos = CMOS(self.cs)
        self.user_request = False
        self.test_offset = 0x38
        self.test_value = 0xAA

    def is_supported(self) -> bool:
        if self.cs.is_core() or (self.cs.Cfg.get_chipset_code() == CHIPSET_CODE_AVN):
            if self.cs.register.is_defined('RC'):
                return True
            self.logger.log_important('RC register not defined for platform.  Skipping module.')
        else:
            self.logger.log_important('Not a Core platform.  Skipping check.')
        return False

    def check_rtclock(self) -> int:
        ll = ul = 0
        check_config_regs = self.cs.register.read('RC') != 0xFFFFFFFF

        if check_config_regs:
            rc_reg = self.cs.register.read('RC')
            self.cs.register.print('RC', rc_reg)
            ll = self.cs.register.get_field('RC', rc_reg, 'LL')
            ul = self.cs.register.get_field('RC', rc_reg, 'UL')
        elif self.user_request:
            self.logger.log_important('Writing to CMOS to determine write protection (original values will be restored)')

            original_val = self.cmos.read_cmos_low(self.test_offset)
            self.cmos.write_cmos_low(self.test_offset, original_val ^ self.test_value)
            if original_val == self.cmos.read_cmos_low(self.test_offset):
                ll = 1
            else:
                self.logger.log_important('Restoring original value')
                self.cmos.write_cmos_low(self.test_offset, original_val)

            original_val = self.cmos.read_cmos_high(self.test_offset)
            self.cmos.write_cmos_high(self.test_offset, original_val ^ self.test_value)
            if original_val == self.cmos.read_cmos_high(self.test_offset):
                ul = 1
            else:
                self.logger.log_important('Restoring original value')
                self.cmos.write_cmos_high(self.test_offset, original_val)
        else:
            self.logger.log_important('Unable to test lock bits without attempting to modify CMOS.')
            self.logger.log('[*] Run chipsec_main manually with the following commandline flags.')
            self.logger.log('[*] python chipsec_main -m common.rtclock -a modify')
            self.result.setStatusBit(self.result.status.VERIFY)
            return self.result.getReturnCode(ModuleResult.WARNING)

        if ll == 1:
            self.logger.log_good('Protected bytes (0x38-0x3F) in low 128-byte bank of RTC memory are locked')
        else:
            self.logger.log_bad('Protected bytes (0x38-0x3F) in low 128-byte bank of RTC memory are not locked')
        if ul == 1:
            self.logger.log_good('Protected bytes (0x38-0x3F) in high 128-byte bank of RTC memory are locked')
        else:
            self.logger.log_bad('Protected bytes (0x38-0x3F) in high 128-byte bank of RTC memory are not locked')

        if (ll == 1) and (ul == 1):
            res = ModuleResult.PASSED
            self.logger.log_passed('Protected locations in RTC memory are locked')
        else:
            res = ModuleResult.WARNING
            self.result.setStatusBit(self.result.status.POTENTIALLY_VULNERABLE)
            self.logger.log_warning('Protected locations in RTC memory are accessible (BIOS may not be using them)')

        return self.result.getReturnCode(res)

    def run(self, module_argv: List[str]) -> int:
        self.logger.start_test('Protected RTC memory locations')

        if len(module_argv) >= 1:
            if module_argv[0].lower() == 'modify':
                self.user_request = True
        self.res = self.check_rtclock()
        return self.res



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2023, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


from unittest.mock import Mock
from typing import Tuple, List
import chipsec.helper.replay.replayhelper as rph
from chipsec_util import ChipsecUtil, parse_args
import chipsec.library.logger


def run_chipsec_util(csu: ChipsecUtil, util_replay_file: str) -> int:
    csu._cs.init(csu._platform, csu._pch, csu._helper, not csu._no_driver, csu._load_config, csu._ignore_platform)
    if util_replay_file:
        csu._helper.config_file = util_replay_file
        csu._helper._load()
    comm = csu.commands[csu._cmd](csu._cmd_args, cs=csu._cs)
    comm.parse_arguments()
    comm.set_up()
    comm.run()
    comm.tear_down()
    return comm.ExitCode


def setup_run_destroy_util_get_log_output(init_replay_file: str, util_name: str, util_args: str = "", util_replay_file: str = "", logging_fucntions_to_capture: List = ['log']) -> Tuple[int, str]:
    chipsec.library.logger._logger.remove_chipsec_logger()
    chipsec.library.logger._logger = Mock()
    chipsec.library.logger._logger.VERBOSE = False
    chipsec.library.logger._logger.DEBUG = False
    chipsec.library.logger._logger.HAL = False
    arg_str = f" {util_args}" if util_args else ""
    cli_cmds = f"{util_name}{arg_str}".strip().split(' ')
    par = parse_args(cli_cmds)
    csu = ChipsecUtil(par, cli_cmds)
    replayHelper = rph.ReplayHelper(init_replay_file)
    csu._helper = replayHelper
    retval = run_chipsec_util(csu, util_replay_file)
    logger_calls = []
    for func in logging_fucntions_to_capture:
        if hasattr(chipsec.library.logger._logger, func):
            logger_calls += getattr(chipsec.library.logger._logger, func).mock_calls
    chipsec.library.logger._logger = chipsec.library.logger.Logger()
    return retval, " ".join([call.args[0] for call in logger_calls])


def setup_run_destroy_util(init_replay_file: str, util_name: str, util_args: str = "", util_replay_file: str = "") -> int:
    retval, _ = setup_run_destroy_util_get_log_output(init_replay_file, util_name, util_args, util_replay_file)
    return retval



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2024, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


from chipsec.library.logger import logger
from typing import List, Optional, Set, Union


class Lock:
    def __init__(self, cs):
        self.cs = cs

    def get(self, lock_name: str, cpu_thread: int = 0, with_print: bool = False, bus: Optional[int] = None) -> Union[int, List[int]]:
        """Retrieves information for the lock associated with the register/field by lock_name."""
        lock = self.cs.Cfg.LOCKS[lock_name]
        reg = lock['register']
        field = lock['field']
        if bus is None:
            reg_data = self.cs.register.read_all(reg, cpu_thread)
        else:
            reg_data = self.cs.register.read(reg, cpu_thread, bus)
            reg_data = [reg_data]
        if logger().VERBOSE or with_print:
            if reg_data:
                for rd in reg_data:
                    self.cs.register.print(reg, rd)
            else:
                logger().log('Register has no data')
        if reg_data:
            return self.cs.register.get_field_all(reg, reg_data, field)
        return reg_data

    def set(self, lock_name: str, lock_value: int, cpu_thread: int = 0, bus: Optional[int] = None) -> bool:
        """Sets the value of a lock associated with the given lock_name."""
        lock = self.cs.Cfg.LOCKS[lock_name]
        reg = lock['register']
        field = lock['field']
        if bus is None:
            reg_data = self.cs.register.read_all(reg, cpu_thread)
            reg_data = self.cs.register.set_field_all(reg, reg_data, field, lock_value)
            return self.cs.register.write_all(reg, reg_data, cpu_thread)
        else:
            reg_data = self.cs.register.read(reg, cpu_thread, bus)
            reg_data = self.cs.register.set_field(reg, reg_data, field, lock_value)
            return self.cs.register.write(reg, reg_data, cpu_thread, bus)

    def is_defined(self, lock_name: str) -> bool:
        """Checks if lock is defined in the XML config"""
        return lock_name in self.cs.Cfg.LOCKS.keys()

    def get_value(self, lock_name: str) -> int:
        """Retrieves the expected value of a lock associated with the lock_name."""
        if logger().DEBUG:
            logger().log(f'Retrieve value for lock {lock_name}')
        return self.cs.Cfg.LOCKS[lock_name]['value']

    def get_desc(self, lock_name: str) -> str:
        """Retrieves the description of a lock assoicated with the lock_name."""
        return self.cs.Cfg.LOCKS[lock_name]['desc']

    def get_type(self, lock_name: str) -> str:
        """Fetcheth the type of a register associated with the lock_name."""
        if 'type' in self.cs.Cfg.LOCKS[lock_name].keys():
            mtype = self.cs.Cfg.LOCKS[lock_name]['type']
        else:
            mtype = 'RW/L'
        return mtype

    def get_list(self) -> List[str]:
        """Retrieve a list of locks that are currently loaded from config files."""
        return list(self.cs.Cfg.LOCKS.keys())

    def get_mask(self, lock_name: str) -> int:
        """Retrieve the field mask of a register associated with the lock_name."""
        lock = self.cs.Cfg.LOCKS[lock_name]
        reg = lock['register']
        field = lock['field']
        return self.cs.register.get_field_mask(reg, field)

    def get_lockedby(self, lock_name: str) -> Optional[List[Set[str]]]:
        """Retrieve a list of registers locked by lock_name."""
        if lock_name in self.cs.Cfg.LOCKEDBY.keys():
            return self.cs.Cfg.LOCKEDBY[lock_name]
        else:
            return None



# CHIPSEC: Platform Security Assessment Framework
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
import os.path

from tests.hardware import test_generic


class GenericUbuntuTest(test_generic.GenericHardwareTest):

    SYSTEM = 'Linux'
    DIST = ('Ubuntu', '16.04', 'xenial')

    PRODUCT_NAME_PATH = "/sys/class/dmi/id/product_name"
    BIOS_VERSION_PATH = "/sys/class/dmi/id/bios_version"
    BOOT_MODE_PATH = "/sys/firmware/efi"

    def product_name(self):
        try:
            product_name = open(self.PRODUCT_NAME_PATH).read().strip()
            return product_name
        except IOError:
            return None

    def bios_version(self):
        try:
            bios_version = open(self.BIOS_VERSION_PATH).read().strip()
            return bios_version
        except IOError:
            return None

    def boot_mode(self):
        """Check if the current boot method is UEFI or Legacy"""
        if os.path.isdir(self.BOOT_MODE_PATH):
            return self.BOOT_MODE_UEFI
        else:
            return self.BOOT_MODE_LEGACY



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2022, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


"""
Tool to test for 'TE Header' vulnerability in Secure Boot implementations as described in
`All Your Boot Are Belong To Us <http://www.c7zero.info/stuff/AllYourBoot_csw14-intel-final.pdf>`_

Usage:
  ``chipsec_main.py -m tools.secureboot.te [-a <mode>,<cfg_file>,<efi_file>]``
      - ``<mode>``

          * ``generate_te``     (default) convert PE EFI binary ``<efi_file>`` to TE binary
          * ``replace_bootloader``  replace bootloader files listed in ``<cfg_file>`` on ESP with modified ``<efi_file>``
          * ``restore_bootloader``  restore original bootloader files from ``.bak`` files

      - ``<cfg_file>``  path to config file listing paths to bootloader files to replace
      - ``<efi_file>``  path to EFI binary to convert to TE binary. If no file path is provided, the tool will look for Shell.efi

Examples:

Convert Shell.efi PE/COFF EFI executable to TE executable:

  ``chipsec_main.py -m tools.secureboot.te -a generate_te,Shell.efi``

Replace bootloaders listed in te.cfg file with TE version of Shell.efi executable:

  ``chipsec_main.py -m tools.secureboot.te -a replace_bootloader,te.cfg,Shell.efi``

Restore bootloaders listed in te.cfg file:

  ``chipsec_main.py -m tools.secureboot.te -a restore_bootloader,te.cfg``

"""

import os
import shutil
import struct
import sys

from chipsec.module_common import BaseModule, SECUREBOOT
from chipsec.library.returncode import ModuleResult
from chipsec.library.logger import logger

TAGS = [SECUREBOOT]
METADATA_TAGS = ['OPENSOURCE', 'IA', 'TOOLS', 'SECUREBOOT', 'TE']


DEFAULT_PE_FILE_PATH = "chipsec/modules/tools/secureboot/Shell.efi"
DEFAULT_CONFIG_FILE_PATH = 'chipsec/modules/tools/secureboot/te.cfg'

# typedef struct _IMAGE_DOS_HEADER
# {
#      WORD e_magic;
#      WORD e_cblp;
#      WORD e_cp;
#      WORD e_crlc;
#      WORD e_cparhdr;
#      WORD e_minalloc;
#      WORD e_maxalloc;
#      WORD e_ss;
#      WORD e_sp;
#      WORD e_csum;
#      WORD e_ip;
#      WORD e_cs;
#      WORD e_lfarlc;
#      WORD e_ovno;
#      WORD e_res[4];
#      WORD e_oemid;
#      WORD e_oeminfo;
#      WORD e_res2[10];
#      LONG e_lfanew;
# } IMAGE_DOS_HEADER, *PIMAGE_DOS_HEADER;

IMAGE_DOS_HEADER = "<14H4HHH10Hi"
IMAGE_DOS_HEADER_size = struct.calcsize(IMAGE_DOS_HEADER)
E_MAGIC = 0x5A4D
E_MAGIC_STR = "MZ"

# typedef struct _IMAGE_DATA_DIRECTORY
# {
#      ULONG VirtualAddress;
#      ULONG Size;
# } IMAGE_DATA_DIRECTORY, *PIMAGE_DATA_DIRECTORY;

IMAGE_DATA_DIRECTORY = "<II"
IMAGE_DATA_DIRECTORY_size = struct.calcsize(IMAGE_DATA_DIRECTORY)

IMAGE_NUMBEROF_DIRECTORY_ENTRIES = 16

# typedef struct _IMAGE_OPTIONAL_HEADER
# {
#      WORD Magic;
#      UCHAR MajorLinkerVersion;
#      UCHAR MinorLinkerVersion;
#      ULONG SizeOfCode;
#      ULONG SizeOfInitializedData;
#      ULONG SizeOfUninitializedData;
#      ULONG AddressOfEntryPoint;
#      ULONG BaseOfCode;
#      ULONG BaseOfData;
#      ULONG ImageBase;
#      ULONG SectionAlignment;
#      ULONG FileAlignment;
#      WORD MajorOperatingSystemVersion;
#      WORD MinorOperatingSystemVersion;
#      WORD MajorImageVersion;
#      WORD MinorImageVersion;
#      WORD MajorSubsystemVersion;
#      WORD MinorSubsystemVersion;
#      ULONG Win32VersionValue;
#      ULONG SizeOfImage;
#      ULONG SizeOfHeaders;
#      ULONG CheckSum;
#      WORD Subsystem;
#      WORD DllCharacteristics;
#      ULONG SizeOfStackReserve;
#      ULONG SizeOfStackCommit;
#      ULONG SizeOfHeapReserve;
#      ULONG SizeOfHeapCommit;
#      ULONG LoaderFlags;
#      ULONG NumberOfRvaAndSizes;
#      IMAGE_DATA_DIRECTORY DataDirectory[16];
# } IMAGE_OPTIONAL_HEADER, *PIMAGE_OPTIONAL_HEADER;

IMAGE_OPTIONAL_HEADER = "<HBB9I6H4I2H6I"
IMAGE_OPTIONAL_HEADER_size = struct.calcsize(IMAGE_OPTIONAL_HEADER)
IMAGE_NT_OPTIONAL_HDR32_MAGIC = 0x10b

# typedef struct {
#   //
#   // Standard fields.
#   //
#   UINT16                    Magic;
#   UINT8                     MajorLinkerVersion;
#   UINT8                     MinorLinkerVersion;
#   UINT32                    SizeOfCode;
#   UINT32                    SizeOfInitializedData;
#   UINT32                    SizeOfUninitializedData;
#   UINT32                    AddressOfEntryPoint;
#   UINT32                    BaseOfCode;
#   //
#   // NT additional fields.
#   //
#   UINT64                    ImageBase;
#   UINT32                    SectionAlignment;
#   UINT32                    FileAlignment;
#   UINT16                    MajorOperatingSystemVersion;
#   UINT16                    MinorOperatingSystemVersion;
#   UINT16                    MajorImageVersion;
#   UINT16                    MinorImageVersion;
#   UINT16                    MajorSubsystemVersion;
#   UINT16                    MinorSubsystemVersion;
#   UINT32                    Win32VersionValue;
#   UINT32                    SizeOfImage;
#   UINT32                    SizeOfHeaders;
#   UINT32                    CheckSum;
#   UINT16                    Subsystem;
#   UINT16                    DllCharacteristics;
#   UINT64                    SizeOfStackReserve;
#   UINT64                    SizeOfStackCommit;
#   UINT64                    SizeOfHeapReserve;
#   UINT64                    SizeOfHeapCommit;
#   UINT32                    LoaderFlags;
#   UINT32                    NumberOfRvaAndSizes;
#   EFI_IMAGE_DATA_DIRECTORY  DataDirectory[EFI_IMAGE_NUMBER_OF_DIRECTORY_ENTRIES];
# } EFI_IMAGE_OPTIONAL_HEADER64;

IMAGE_OPTIONAL_HEADER64 = "<HBBIIIIIQIIHHHHHHIIIIHHQQQQII"
IMAGE_OPTIONAL_HEADER64_size = struct.calcsize(IMAGE_OPTIONAL_HEADER64)
IMAGE_NT_OPTIONAL_HDR64_MAGIC = 0x20b

# typedef struct _IMAGE_FILE_HEADER
# {
#      WORD Machine;
#      WORD NumberOfSections;
#      ULONG TimeDateStamp;
#      ULONG PointerToSymbolTable;
#      ULONG NumberOfSymbols;
#      WORD SizeOfOptionalHeader;
#      WORD Characteristics;
# } IMAGE_FILE_HEADER, *PIMAGE_FILE_HEADER;

IMAGE_FILE_HEADER = "<2H3I2H"
IMAGE_FILE_HEADER_size = struct.calcsize(IMAGE_FILE_HEADER)

# typedef struct _IMAGE_NT_HEADERS
# {
#      ULONG Signature;
#      IMAGE_FILE_HEADER FileHeader;
#      IMAGE_OPTIONAL_HEADER OptionalHeader;
# } IMAGE_NT_HEADERS, *PIMAGE_NT_HEADERS;

IMAGE_NT_SIGNATURE = 0x00004550
IMAGE_NT_HEADERS_size = (4 + IMAGE_FILE_HEADER_size + IMAGE_OPTIONAL_HEADER_size + IMAGE_NUMBEROF_DIRECTORY_ENTRIES * IMAGE_DATA_DIRECTORY_size)

# typedef struct _IMAGE_SECTION_HEADER
# {
#      UCHAR Name[8];
#      ULONG Misc;
#      ULONG VirtualAddress;
#      ULONG SizeOfRawData;
#      ULONG PointerToRawData;
#      ULONG PointerToRelocations;
#      ULONG PointerToLinenumbers;
#      WORD NumberOfRelocations;
#      WORD NumberOfLinenumbers;
#      ULONG Characteristics;
# } IMAGE_SECTION_HEADER, *PIMAGE_SECTION_HEADER;

IMAGE_SECTION_HEADER = "<8s6I2HI"
IMAGE_SECTION_HEADER_size = struct.calcsize(IMAGE_SECTION_HEADER)

# PE executable structure
#
#   MS-DOS header
#     ...
#     e_lfanew -------------------+
#                                 |
#                                 |
#   IMAGE_NT_HEADERS Header  <----+
#    ...
#   SECTION TABLE
#    ...
# TE header
#
# typedef struct {
#   UINT16                    Signature;            // signature for TE format = "VZ"
#   UINT16                    Machine;              // from the original file header
#   UINT8                     NumberOfSections;     // from the original file header
#   UINT8                     Subsystem;            // from original optional header
#   UINT16                    StrippedSize;         // how many bytes we removed from the header
#   UINT32                    AddressOfEntryPoint;  // offset to entry point -- from original optional header
#   UINT32                    BaseOfCode;           // from original image -- required for ITP debug
#   UINT64                    ImageBase;            // from original file header
#   IMAGE_DATA_DIRECTORY      DataDirectory[2];     // only base relocation and debug directory
# } EFI_TE_IMAGE_HEADER;

EFI_TE_IMAGE_HEADER = "<HHBBHIIQ"
EFI_TE_IMAGE_HEADER_SIGNATURE = 0x5A56

EFI_IMAGE_DIRECTORY_ENTRY_EXPORT = 0
EFI_IMAGE_DIRECTORY_ENTRY_IMPORT = 1
EFI_IMAGE_DIRECTORY_ENTRY_RESOURCE = 2
EFI_IMAGE_DIRECTORY_ENTRY_EXCEPTION = 3
EFI_IMAGE_DIRECTORY_ENTRY_SECURITY = 4
EFI_IMAGE_DIRECTORY_ENTRY_BASERELOC = 5
EFI_IMAGE_DIRECTORY_ENTRY_DEBUG = 6
EFI_IMAGE_DIRECTORY_ENTRY_COPYRIGHT = 7
EFI_IMAGE_DIRECTORY_ENTRY_GLOBALPTR = 8
EFI_IMAGE_DIRECTORY_ENTRY_TLS = 9
EFI_IMAGE_DIRECTORY_ENTRY_LOAD_CONFIG = 10

EFI_TE_IMAGE_DIRECTORY_ENTRY_BASERELOC = 0
EFI_TE_IMAGE_DIRECTORY_ENTRY_DEBUG = 1


def IsValidPEHeader(data):
    size = len(data)
    if size < IMAGE_DOS_HEADER_size:
        return False
    signature, = struct.unpack("<H", data[:2])
    if signature != E_MAGIC:
        return False
    e_lfanew, = struct.unpack("<I", data[IMAGE_DOS_HEADER_size - 4:IMAGE_DOS_HEADER_size])
    if e_lfanew >= size:
        return False
    if (size - e_lfanew) < IMAGE_NT_HEADERS_size:
        return False
    pe_signature, = struct.unpack("<I", data[e_lfanew:e_lfanew + 4])
    if pe_signature != IMAGE_NT_SIGNATURE:
        return False
    return True


def replace_header(data):
    if not IsValidPEHeader(data):
        return None
    size = len(data)
    e_lfanew, = struct.unpack("<I", data[IMAGE_DOS_HEADER_size - 4:IMAGE_DOS_HEADER_size])
    #                          TimeDateStamp, PointerToSymbolTable, NumberOfSymbols, SizeOfOptionalHeader, Characteristics
    Machine, NumberOfSections, u1, u2, u3, SizeOfOptionalHeader, u5 \
        = struct.unpack(IMAGE_FILE_HEADER, data[e_lfanew + 4:e_lfanew + 4 + IMAGE_FILE_HEADER_size])
    StrippedSize = e_lfanew + 4 + IMAGE_FILE_HEADER_size + SizeOfOptionalHeader
    if StrippedSize > size:
        return None
    if StrippedSize & ~0xffff:
        return None
    dof = e_lfanew + 4 + IMAGE_FILE_HEADER_size
    Magic, = struct.unpack("<H", data[dof:dof + 2])
    if Magic == IMAGE_NT_OPTIONAL_HDR32_MAGIC:
        Magic, MajorLinkerVersion, MinorLinkerVersion, SizeOfCode, SizeOfInitializedData, \
            SizeOfUninitializedData, AddressOfEntryPoint, BaseOfCode, BaseOfData, ImageBase,  \
            SectionAlignment, FileAlignment, MajorOperatingSystemVersion, MinorOperatingSystemVersion, \
            MajorImageVersion, MinorImageVersion, MajorSubsystemVersion, MinorSubsystemVersion, \
            Win32VersionValue, SizeOfImage, SizeOfHeaders, CheckSum, Subsystem, DllCharacteristics, \
            SizeOfStackReserve, SizeOfStackCommit, SizeOfHeapReserve, SizeOfHeapCommit, LoaderFlags, \
            NumberOfRvaAndSizes = struct.unpack(IMAGE_OPTIONAL_HEADER, data[dof:dof + IMAGE_OPTIONAL_HEADER_size])
        dof = dof + IMAGE_OPTIONAL_HEADER_size
    elif Magic == IMAGE_NT_OPTIONAL_HDR64_MAGIC:
        Magic, MajorLinkerVersion, MinorLinkerVersion, SizeOfCode, SizeOfInitializedData, \
            SizeOfUninitializedData, AddressOfEntryPoint, BaseOfCode, ImageBase, SectionAlignment, \
            FileAlignment, MajorOperatingSystemVersion, MinorOperatingSystemVersion, MajorImageVersion, \
            MinorImageVersion, MajorSubsystemVersion, MinorSubsystemVersion, Win32VersionValue, \
            SizeOfImage, SizeOfHeaders, CheckSum, Subsystem, DllCharacteristics, SizeOfStackReserve, \
            SizeOfStackCommit, SizeOfHeapReserve, SizeOfHeapCommit, LoaderFlags, NumberOfRvaAndSizes \
            = struct.unpack(IMAGE_OPTIONAL_HEADER64, data[dof:dof + IMAGE_OPTIONAL_HEADER64_size])
        dof = dof + IMAGE_OPTIONAL_HEADER64_size
    else:
        return None
    if NumberOfSections & ~0xFF:
        return None
    if Subsystem & ~0xFF:
        return None

    basereloc_off = dof + EFI_IMAGE_DIRECTORY_ENTRY_BASERELOC * IMAGE_DATA_DIRECTORY_size
    debug_off = dof + EFI_IMAGE_DIRECTORY_ENTRY_DEBUG * IMAGE_DATA_DIRECTORY_size
    BASERELOC = "\x00\x00\x00\x00\x00\x00\x00\x00"
    DEBUG = "\x00\x00\x00\x00\x00\x00\x00\x00"
    if NumberOfRvaAndSizes > EFI_IMAGE_DIRECTORY_ENTRY_BASERELOC:
        BASERELOC = data[basereloc_off:basereloc_off + IMAGE_DATA_DIRECTORY_size]
    if NumberOfRvaAndSizes > EFI_IMAGE_DIRECTORY_ENTRY_DEBUG:
        DEBUG = data[debug_off:debug_off + IMAGE_DATA_DIRECTORY_size]
    te_header = struct.pack(EFI_TE_IMAGE_HEADER,
                            EFI_TE_IMAGE_HEADER_SIGNATURE, Machine, NumberOfSections, Subsystem, StrippedSize, AddressOfEntryPoint, BaseOfCode, ImageBase)
    te_data = te_header + BASERELOC + DEBUG + data[StrippedSize:]
    return te_data


def produce_te(fname, outfname):
    data = ''
    with open(fname, 'rb') as f:
        data = f.read()
    te_data = replace_header(data)
    if te_data is None:
        return 0
    with open(outfname, 'wb') as fte:
        fte.write(te_data)
    return 1


def replace_efi_binary(orig_efi_binary, new_efi_binary):
    logger().log(f'[*] Replacing EFI binary \'{orig_efi_binary}\'..')
    te_binary = new_efi_binary + '.te'
    if not os.path.exists(te_binary):
        produce_te(new_efi_binary, te_binary)
    # back up original binary
    backup = orig_efi_binary + '.bak'
    if not os.path.exists(backup):
        os.rename(orig_efi_binary, backup)
    try:
        shutil.copy(te_binary, orig_efi_binary)
    except OSError as err:
        logger().log_error(f'Cannot replace binary ({err})')
        return False
    return True


def umount(drive):
    import subprocess
    if os.path.exists(drive):
        res = subprocess.call(["mountvol.exe", drive, "/D"])
        if res != 0:
            logger().log_warning(f'Cannot unmount EFI System partition: {res:d}')


def get_efi_mount():
    import subprocess
    for l in range(ord('z'), ord('a'), -1):
        if not os.path.exists('%c:\\' % l):
            res = subprocess.call(["mountvol.exe", "%c:\\" % l, "/S"])
            if res != 0:
                logger().log_error(f'Cannot mount EFI System partition (status = {res:d})')
                return None
            return '%c:\\' % l
    logger().log_error("Cannot mount EFI System partition. No drive letters to use.")
    return None


def get_bootloader_paths(cfg_file):
    bootloader_paths = []
    with open(cfg_file, 'r') as fcfg:
        logger().log(f'[*] reading paths from \'{cfg_file}\'..')
        for line in fcfg:
            bl_path = line.rstrip()
            if bl_path is not None:
                logger().log(f'    adding path \'{bl_path}\'..')
                bootloader_paths.append(bl_path)
    return bootloader_paths


def replace_bootloader(bootloader_paths, new_bootloader_file, do_mount=True):
    logger().log("[*] Replacing bootloaders on EFI System Partition (ESP)...")
    dsk = get_efi_mount() if do_mount else ''
    if dsk is None:
        return False
    try:
        for pth in bootloader_paths:
            bootloader_path = os.path.join(dsk, pth)
            if os.path.exists(bootloader_path):
                replace_efi_binary(bootloader_path, new_bootloader_file)
            else:
                logger().log_warning(f'Bootloader {bootloader_path} does not exist on ESP')
    finally:
        if do_mount:
            umount(dsk)
    logger().log("[*] You will need to reboot the system to see the changes")
    return True


def restore_efi_binary(orig_efi_binary):
    logger().log(f'[*] Restoring {orig_efi_binary}..')
    backup = orig_efi_binary + ".bak"
    if not os.path.exists(backup):
        logger().log_error(f'Cannot restore original binary: \'{backup}\' not found')
        return False
    try:
        if os.path.exists(orig_efi_binary):
            os.remove(orig_efi_binary)
        os.rename(backup, orig_efi_binary)
    except OSError as err:
        logger().log_error(f'Cannot restore original binary ({err})')
        return False
    return True


def restore_bootloader(bootloader_paths, do_mount=True):
    logger().log("[*] Restoring bootloaders on EFI System Partition (ESP)...")
    dsk = get_efi_mount() if do_mount else ''
    if dsk is None:
        return False
    for pth in bootloader_paths:
        bootloader_path = os.path.join(dsk, pth)
        if os.path.exists(bootloader_path):
            restore_efi_binary(bootloader_path)
    if do_mount:
        umount(dsk)
    logger().log("[*] You will need to reboot the system to see the changes")
    return True


def confirm():
    logger().log_important("***************************************************************************************")
    logger().log_important("*")
    logger().log_important("* RUNNING THIS TOOL MAY RESULT IN UNBOOTABLE OS!")
    logger().log_important("* USE IT FOR TESTING PURPOSES ON TEST SYSTEMS ONLY")
    logger().log_important("*")
    logger().log_important("* The tool converts PE/COFF EFI executables to TE EFI executables.")
    logger().log_important("* The tool can also automatically replace files (boot loaders)")
    logger().log_important("* listed in the configuration file with the generated TE executable.")
    logger().log_important("*")
    logger().log_important("* If after reboot, TE executable runs then the firmware doesn't properly")
    logger().log_important("* enforce Secure Boot checks on TE EFI executables")
    logger().log_important("*")
    logger().log_important("* If TE executable doesn't run then the firmware correctly blocked it.")
    logger().log_important("* To restore OS boot loader in this case you may use one of the following:")
    logger().log_important("* - Disable Secure Boot in BIOS, boot to external drive (e.g. Linux or UEFI shell)")
    logger().log_important("*   then restore original boot loader executables from .bak files")
    logger().log_important("* - On Windows, use recovery mode which should automatically restore correct executables")
    logger().log_important("*")
    logger().log_important("***************************************************************************************")
    s = input("Type 'yes' to continue running the tool > ")
    if s.lower() not in ['yes', 'y']:
        sys.exit(0)


def usage():
    logger().log('Usage:\n' +
                 'chipsec_main.py -m tools.secureboot.te [-a <mode>,<cfg_file>,<efi_file>]\n' +
                 '    <mode>\n' +
                 '      generate_te        - (default) convert PE EFI binary <efi_file> to TE binary\n' +
                 '      replace_bootloader - replace bootloader files listed in <cfg_file> on ESP with modified <efi_file>\n' +
                 '      restore_bootloader - restore original bootloader files from .bak files\n' +
                 '    <cfg_file>           - path to config file listing paths to bootloader files to replace\n' +
                 '    <efi_file>           - path to EFI binary to convert to TE binary\n' +
                 '                           If no file path is provided, the tool will look for Shell.efi\n')


class te(BaseModule):

    def __init__(self):
        BaseModule.__init__(self)

    def is_supported(self):
        # win8 = self.cs.helper.is_win8_or_greater()
        efi_mode = self.cs.helper.EFI_supported()
        if not efi_mode:
            self.logger.log_not_applicable("OS did not boot in UEFI mode")
        return efi_mode

    def run(self, module_argv):
        self.logger.start_test("'TE Header' Secure Boot Bypass Test")
        usage()

        sts = False
        do_mount = True
        file_path = DEFAULT_PE_FILE_PATH
        te_cfg = DEFAULT_CONFIG_FILE_PATH
        mode = module_argv[0] if len(module_argv) > 0 else 'generate_te'

        if 'generate_te' == mode:
            if len(module_argv) > 1:
                file_path = module_argv[1]
            if not os.path.exists(file_path):
                self.logger.log_error(f'Cannot find file \'{file_path}\'')
                self.logger.log_error(f'Please download it from https://github.com/chipsec/chipsec/releases/download/binaries/Shell.efi ')
                self.logger.log_error(f'And move file to \'{file_path}\'')
                self.result.setStatusBit(self.result.status.ACCESS_RW)
                return self.result.getReturnCode(ModuleResult.ERROR)

            sts = replace_efi_binary(file_path, file_path)

        elif ('restore_bootloader' == mode) or ('replace_bootloader' == mode):
            confirm()

            if len(module_argv) > 1:
                te_cfg = module_argv[1]
            if not os.path.exists(te_cfg):
                self.logger.log_error(f'Cannot find file \'{te_cfg}\'')
                self.result.setStatusBit(self.result.status.ACCESS_RW)
                return self.result.getReturnCode(ModuleResult.ERROR)

            bootloader_paths = get_bootloader_paths(te_cfg)
            if len(bootloader_paths) == 0:
                self.logger.log("[*] no bootloaders to replace. Exit...")
                self.result.setStatusBit(self.result.status.FEATURE_DISABLED)
                return self.result.getReturnCode(ModuleResult.WARNING)

            do_mount = self.cs.os_helper.is_windows()  # @TODO
            if 'restore_bootloader' == mode:
                sts = restore_bootloader(bootloader_paths, do_mount)
            elif 'replace_bootloader' == mode:
                if len(module_argv) > 2:
                    file_path = module_argv[2]
                sts = replace_bootloader(bootloader_paths, file_path, do_mount)

        else:
            self.logger.log_error(f'Invalid mode: \'{mode}\'')

        if sts:
            self.result.setStatusBit(self.result.status.SUCCESS)
            return self.result.getReturnCode(ModuleResult.PASSED)
        else:
            self.result.setStatusBit(self.result.status.RESTORE)
            return self.result.getReturnCode(ModuleResult.ERROR)



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


"""
Checks protections of the S3 resume boot-script implemented by the UEFI based firmware

References:

`VU#976132 UEFI implementations do not properly secure the EFI S3 Resume Boot Path boot script <https://www.kb.cert.org/vuls/id/976132>`_

`Technical Details of the S3 Resume Boot Script Vulnerability <http://www.intelsecurity.com/advanced-threat-research/content/WP_Intel_ATR_S3_ResBS_Vuln.pdf>`_ by Intel Security's Advanced Threat Research team.

`Attacks on UEFI Security <https://events.ccc.de/congress/2014/Fahrplan/system/attachments/2557/original/AttacksOnUEFI_Slides.pdf>`_ by Rafal Wojtczuk and Corey Kallenberg.

`Attacking UEFI Boot Script <https://bromiumlabs.files.wordpress.com/2015/01/venamis_whitepaper.pdf>`_ by Rafal Wojtczuk and Corey Kallenberg.

`Exploiting UEFI boot script table vulnerability <http://blog.cr4.sh/2015/02/exploiting-uefi-boot-script-table.html>`_ by Dmytro Oleksiuk.

Usage:
    ``chipsec_main.py -m common.uefi.s3bootscript [-a <script_address>]``

    - ``-a <script_address>``: Specify the bootscript address

Where:
    - ``[]``: optional line

Examples:
    >>> chipsec_main.py -m common.uefi.s3bootscript
    >>> chipsec_main.py -m common.uefi.s3bootscript -a 0x00000000BDE10000

.. NOTE::
    Requires an OS with UEFI Runtime API support.
"""

from chipsec.module_common import BaseModule, IA, BIOS, SMM, SECUREBOOT, ARCHIVED
from chipsec.library.returncode import ModuleResult
from chipsec.library.defines import BOUNDARY_1MB, BOUNDARY_4GB
from chipsec.hal.uefi import UEFI, parse_script
from chipsec.hal.uefi_common import S3BootScriptOpcode, S3BOOTSCRIPT_ENTRY
from typing import List

TAGS = [IA, BIOS, SMM, SECUREBOOT, ARCHIVED]
METADATA_TAGS = ['OPENSOURCE', 'IA', 'COMMON', 'UEFI', 'S3BOOTSCRIPT', 'ARCHIVED']

########################################################################################################
#
# Main module functionality
#
########################################################################################################
BOOTSCRIPT_OK = 0x0
BOOTSCRIPT_INSIDE_SMRAM = 0x1
BOOTSCRIPT_OUTSIDE_SMRAM = 0x2
DISPATCH_OPCODES_UNPROTECTED = 0x4
DISPATCH_OPCODES_PROTECTED = 0x8

HIGH_BIOS_RANGE_SIZE = 2 * BOUNDARY_1MB


class s3bootscript(BaseModule):

    def __init__(self):
        BaseModule.__init__(self)
        self._uefi = UEFI(self.cs)
        self.result.url = 'https://chipsec.github.io/modules/chipsec.modules.common.uefi.s3bootscript.html'

    def is_supported(self) -> bool:
        supported = self.cs.helper.EFI_supported()
        if not supported:
            self.logger.log("OS does not support UEFI Runtime API")
        return supported

    def is_inside_SMRAM(self, pa: int) -> bool:
        return (pa >= self.smrambase and pa < self.smramlimit)

    def is_inside_SPI(self, pa: int) -> bool:
        return (pa >= (BOUNDARY_4GB - HIGH_BIOS_RANGE_SIZE) and pa < BOUNDARY_4GB)

    def check_dispatch_opcodes(self, bootscript_entries: List[S3BOOTSCRIPT_ENTRY]) -> bool:
        self.logger.log('[*] Checking entry-points of Dispatch opcodes..')
        dispatch_ep_ok = True
        n_dispatch = 0
        for e in bootscript_entries:
            if e.decoded_opcode is None:
                continue
            if S3BootScriptOpcode.EFI_BOOT_SCRIPT_DISPATCH_OPCODE == e.decoded_opcode.opcode:
                n_dispatch += 1
                dispatchstr = f"Dispatch opcode (off 0x{e.offset_in_script:04X}) with entry-point 0x{e.decoded_opcode.entrypoint:016X}"
                if not self.is_inside_SMRAM(e.decoded_opcode.entrypoint) and not self.is_inside_SPI(e.decoded_opcode.entrypoint):
                    dispatch_ep_ok = False
                    self.logger.log_bad(dispatchstr + " > UNPROTECTED")
                else:
                    self.logger.log_good(dispatchstr + " > PROTECTED")
        self.logger.log(f"[*] Found {n_dispatch:d} Dispatch opcodes")
        return dispatch_ep_ok

    def check_s3_bootscript(self, bootscript_pa: int) -> int:
        res = BOOTSCRIPT_OK
        self.logger.log(f"[*] Checking S3 boot-script at 0x{bootscript_pa:016X}")

        # Checking if it's in SMRAM
        scriptInsideSMRAM = self.is_inside_SMRAM(bootscript_pa)
        if scriptInsideSMRAM:
            res |= BOOTSCRIPT_INSIDE_SMRAM
            self.logger.log_good('S3 boot-script is in SMRAM')
            self.logger.log_important("Note: the test could not verify Dispatch opcodes because the script is in SMRAM. Entry-points of Dispatch opcodes also need to be protected.")
        else:
            res |= BOOTSCRIPT_OUTSIDE_SMRAM
            self.logger.log_bad('S3 boot-script is not in SMRAM')
            self.logger.log('[*] Reading S3 boot-script from memory..')
            script_all = self.cs.mem.read_physical_mem(bootscript_pa, 0x100000)
            self.logger.log('[*] Decoding S3 boot-script opcodes..')
            script_entries = parse_script(script_all, False)
            dispatch_opcodes_ok = self.check_dispatch_opcodes(script_entries)
            if dispatch_opcodes_ok:
                res |= DISPATCH_OPCODES_PROTECTED
                self.logger.log_important("S3 boot-script is not in protected memory but didn't find unprotected Dispatch entry-points")
            else:
                res |= DISPATCH_OPCODES_UNPROTECTED
                self.logger.log_bad('Entry-points of Dispatch opcodes in S3 boot-script are not in protected memory')
        return res

    def check_s3_bootscripts(self, bsaddress=None) -> int:
        res = 0

        if bsaddress:
            bootscript_PAs = [bsaddress]
        else:
            found, bootscript_PAs = self._uefi.find_s3_bootscript()
            if not found:
                self.logger.log_good("Didn't find any S3 boot-scripts in EFI variables")
                self.logger.log_warning("S3 Boot-Script was not found. Firmware may be using other ways to store/locate it, or OS might be blocking access.")
                self.result.setStatusBit(self.result.status.VERIFY)
                return ModuleResult.WARNING

            self.logger.log_important(f'Found {len(bootscript_PAs):d} S3 boot-script(s) in EFI variables')

        for bootscript_pa in bootscript_PAs:
            if 0 == bootscript_pa:
                continue
            res |= self.check_s3_bootscript(bootscript_pa)

        self.logger.log('')

        if (res & BOOTSCRIPT_OUTSIDE_SMRAM) != 0:
            # BOOTSCRIPT_OUTSIDE_SMRAM
            if (res & DISPATCH_OPCODES_UNPROTECTED) != 0:
                # DISPATCH_OPCODES_UNPROTECTED
                status = ModuleResult.FAILED
                self.result.setStatusBit(self.result.status.PROTECTION)
                self.logger.log_failed('S3 Boot-Script and Dispatch entry-points do not appear to be protected')
            else:
                # DISPATCH_OPCODES_PROTECTED
                status = ModuleResult.WARNING
                self.result.setStatusBit(self.result.status.VERIFY)
                self.logger.log_warning('S3 Boot-Script is not in SMRAM but Dispatch entry-points appear to be protected. Recommend further testing')
        else:
            # BOOTSCRIPT_INSIDE_SMRAM
            status = ModuleResult.WARNING
            self.result.setStatusBit(self.result.status.VERIFY)
            self.logger.log_warning("S3 Boot-Script is inside SMRAM. The script is protected but Dispatch opcodes cannot be inspected")

        self.logger.log_important("Additional testing of the S3 boot-script can be done using tools.uefi.s3script_modify")

        return status

    def run(self, module_argv: List[str]) -> int:
        self.logger.start_test("S3 Resume Boot-Script Protections")

        if len(module_argv) > 2:
            self.logger.log_error('Expected module options: -a <bootscript_address>')
            self.result.setStatusBit(self.result.status.UNSUPPORTED_OPTION)
            return self.result.getReturnCode(ModuleResult.ERROR)

        script_pa = None

        if len(module_argv) > 0:
            script_pa = int(module_argv[0], 16)
            self.logger.log(f'[*] Using manually assigned S3 Boot-Script table base: 0x{script_pa:016X}')
        (self.smrambase, self.smramlimit, self.smramsize) = self.cs.cpu.get_SMRAM()
        if (self.smrambase is not None) and (self.smramlimit is not None):
            self.logger.log(f'[*] SMRAM: Base = 0x{self.smrambase:016X}, Limit = 0x{self.smramlimit:016X}, Size = 0x{self.smramsize:08X}')

        try:
            if script_pa is not None:
                self.res = self.check_s3_bootscripts(script_pa)
            else:
                self.res = self.check_s3_bootscripts()
        except:
            self.logger.log_error("The module was not able to recognize the S3 resume boot script on this platform.")
            if self.logger.VERBOSE:
                raise
            self.res = ModuleResult.ERROR

        return self.result.getReturnCode(self.res)



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

"""
Abstracts support for various OS/environments, wrapper around platform specific code that invokes kernel driver
"""

import os
import errno
import importlib
import platform
import traceback
import sys
from typing import List, Any
from chipsec.library.file import get_main_dir, TOOLS_DIR
from chipsec.library.logger import logger
from chipsec.helper.nonehelper import NoneHelper
from chipsec.library.exceptions import OsHelperError


def get_tools_path() -> str:
    return os.path.normpath(os.path.join(get_main_dir(), TOOLS_DIR))


class OsHelper:
    def __init__(self):
        self.avail_helpers = {}
        self.load_helpers()
        self.filecmds = None
        self.helper = self.get_default_helper()
        if (not self.helper):
            os_system = platform.system()
            raise OsHelperError(f'Could not load any helpers for \'{os_system}\' environment (unsupported environment?)', errno.ENODEV)
        else:
            if sys.version[0] == '2':
                logger().log_warning('*****************************************************************************')
                logger().log_warning('* !! Python 2 is deprecated and not supported. Please update to Python 3 !! *')
                logger().log_warning('* !!                           Exiting CHIPSEC                           !! *')
                logger().log_warning('*****************************************************************************')
                sys.exit(0)
            self.os_system = self.helper.os_system
            self.os_release = self.helper.os_release
            self.os_version = self.helper.os_version
            self.os_machine = self.helper.os_machine

    def load_helpers(self) -> None:
        helper_dir = os.path.join(get_main_dir(), 'chipsec', 'helper')
        helpers = [os.path.basename(f) for f in os.listdir(helper_dir)
                   if os.path.isdir(os.path.join(helper_dir, f)) and not os.path.basename(f).startswith('__')]

        for helper in helpers:
            helper_path = ''
            try:
                helper_path = f'chipsec.helper.{helper}.{helper}helper'
                hlpr = importlib.import_module(helper_path)
                self.avail_helpers[f'{helper}helper'] = hlpr
            except ImportError:
                logger().log_debug(f'Unable to load helper: {helper}')

    def get_helper(self, name: str) -> Any:
        ret = None
        if name in self.avail_helpers:
            ret = self.avail_helpers[name].get_helper()
        return ret

    def get_available_helpers(self) -> List[str]:
        return sorted(self.avail_helpers.keys())

    def get_base_helper(self):
        return NoneHelper()

    def get_default_helper(self):
        ret = None
        if self.is_linux():
            ret = self.get_helper('linuxhelper')
        elif self.is_windows():
            ret = self.get_helper('windowshelper')
        elif self.is_efi():
            ret = self.get_helper('efihelper')
        elif self.is_dal():
            ret = self.get_helper('dalhelper')
        if ret is None:
            ret = self.get_base_helper()
        return ret

    def is_dal(self) -> bool:
        return 'itpii' in sys.modules

    def is_efi(self) -> bool:
        return platform.system().lower().startswith('efi') or platform.system().lower().startswith('uefi')

    def is_linux(self) -> bool:
        return 'linux' == platform.system().lower()

    def is_windows(self) -> bool:
        return 'windows' == platform.system().lower()

    def is_win8_or_greater(self) -> bool:
        win8_or_greater = self.is_windows() and (self.os_release.startswith('8') or ('2008Server' in self.os_release) or ('2012Server' in self.os_release))
        return win8_or_greater

    def is_macos(self) -> bool:
        return 'darwin' == platform.system().lower()

    def getcwd(self) -> str:
        return os.getcwd()


_helper = None


def helper():
    global _helper
    if _helper is None:
        try:
            _helper = OsHelper()
        except BaseException as msg:
            if logger().DEBUG:
                logger().log_error(str(msg))
                logger().log_bad(traceback.format_exc())
            raise
    return _helper



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2017, Google
# Copyright (c) 2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""
>>> chipsec_util reg read <reg_name> [<field_name>]
>>> chipsec_util reg read_field <reg_name> <field_name>
>>> chipsec_util reg write <reg_name> <value>
>>> chipsec_util reg write_field <reg_name> <field_name> <value>
>>> chipsec_util reg get_control <control_name>
>>> chipsec_util reg set_control <control_name> <value>

Examples:

>>> chipsec_util reg read SMBUS_VID
>>> chipsec_util reg read HSFC FGO
>>> chipsec_util reg read_field HSFC FGO
>>> chipsec_util reg write SMBUS_VID 0x8088
>>> chipsec_util reg write_field BC BLE 0x1
>>> chipsec_util reg get_control BiosWriteEnable
>>> chipsec_util reg set_control BiosLockEnable 0x1
"""

from chipsec.command import BaseCommand, toLoad
from argparse import ArgumentParser


class RegisterCommand(BaseCommand):

    def requirements(self) -> toLoad:
        return toLoad.All

    def parse_arguments(self) -> None:
        parser = ArgumentParser(prog='chipsec_util reg', usage=__doc__)
        subparsers = parser.add_subparsers()

        parser_read = subparsers.add_parser('read')
        parser_read.add_argument('reg_name', type=str, help='Register name')
        parser_read.add_argument('field_name', type=str, nargs='?', default=None, help='Field name')
        parser_read.set_defaults(func=self.reg_read)

        parser_readfield = subparsers.add_parser('read_field')
        parser_readfield.add_argument('reg_name', type=str, help='Register name')
        parser_readfield.add_argument('field_name', type=str, help='Field name')
        parser_readfield.set_defaults(func=self.reg_read_field)

        parser_write = subparsers.add_parser('write')
        parser_write.add_argument('reg_name', type=str, help='Register name')
        parser_write.add_argument('value', type=lambda x: int(x, 16), help='Value (hex)')
        parser_write.set_defaults(func=self.reg_write)

        parser_writefield = subparsers.add_parser('write_field')
        parser_writefield.add_argument('reg_name', type=str, help='Register name')
        parser_writefield.add_argument('field_name', type=str, help='Field name')
        parser_writefield.add_argument('value', type=lambda x: int(x, 16), help='Value (hex)')
        parser_writefield.set_defaults(func=self.reg_write_field)

        parser_getcontrol = subparsers.add_parser('get_control')
        parser_getcontrol.add_argument('control_name', type=str, help='Control name')
        parser_getcontrol.set_defaults(func=self.reg_get_control)

        parser_setcontrol = subparsers.add_parser('set_control')
        parser_setcontrol.add_argument('control_name', type=str, help='Control name')
        parser_setcontrol.add_argument('value', type=lambda x: int(x, 16), help='Value (hex)')
        parser_setcontrol.set_defaults(func=self.reg_set_control)

        parser.parse_args(self.argv, namespace=self)

    def reg_read(self):
        if self.field_name is not None:
            value = self.cs.register.read_field(self.reg_name, self.field_name)
            self.logger.log("[CHIPSEC] {}.{}=0x{:X}".format(self.reg_name, self.field_name, value))
        else:
            value = self.cs.register.read(self.reg_name)
            self.logger.log("[CHIPSEC] {}=0x{:X}".format(self.reg_name, value))
            self.cs.register.print(self.reg_name, value)

    def reg_read_field(self):
        if self.cs.register.has_field(self.reg_name, self.field_name):
            value = self.cs.register.read_field(self.reg_name, self.field_name)
            self.logger.log("[CHIPSEC] {}.{}=0x{:X}".format(self.reg_name, self.field_name, value))
        else:
            self.logger.log_error("[CHIPSEC] Register '{}' doesn't have field '{}' defined".format(self.reg_name, self.field_name))

    def reg_write(self):
        self.logger.log("[CHIPSEC] Writing {} < 0x{:X}".format(self.reg_name, self.value))
        self.cs.register.write(self.reg_name, self.value)

    def reg_write_field(self):
        if self.cs.register.has_field(self.reg_name, self.field_name):
            self.logger.log("[CHIPSEC] Writing {}.{} < 0x{:X}".format(self.reg_name, self.field_name, self.value))
            self.cs.register.write_field(self.reg_name, self.field_name, self.value)
        else:
            self.logger.log_error("[CHIPSEC] Register '{}' doesn't have field '{}' defined".format(self.reg_name, self.field_name))

    def reg_get_control(self):
        if self.cs.control.is_defined(self.control_name):
            value = self.cs.control.get(self.control_name)
            self.logger.log("[CHIPSEC] {} = 0x{:X}".format(self.control_name, value))
        else:
            self.logger.log_error("[CHIPSEC] Control '{}' isn't defined".format(self.control_name))

    def reg_set_control(self):
        if self.cs.control.is_defined(self.control_name):
            self.cs.control.set(self.control_name, self.value)
            self.logger.log("[CHIPSEC] Setting control {} < 0x{:X}".format(self.control_name, self.value))
        else:
            self.logger.log_error("[CHIPSEC] Control '{}' isn't defined".format(self.control_name))


commands = {'reg': RegisterCommand}



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


"""
Trusted Platform Module (TPM) HAL component

https://trustedcomputinggroup.org
"""

from chipsec.hal import hal_base
from chipsec.hal import acpi
from chipsec.hal import tpm_interface
from chipsec.library.tpm import tpm_defines


class TPM(hal_base.HALBase):
    def __init__(self, cs):
        super(TPM, self).__init__(cs)
        self.helper = cs.helper
        self.tpm_acpi = acpi.ACPI(self.cs)
        self.version = self.read_tpm_version()
        self.interface = self.read_tpm_interface()
        self.tpm = self.init_tpm()

    def read_tpm_version(self) -> str:
        tpm1_acpi_present = self.tpm_acpi.is_ACPI_table_present('TCPA')
        tpm2_acpi_present = self.tpm_acpi.is_ACPI_table_present('TPM2')
        if (not tpm1_acpi_present and not tpm2_acpi_present):
            raise RuntimeError('No TPM recognized')  # TODO: this might need a proper error
        elif tpm1_acpi_present and not tpm2_acpi_present:
            return tpm_defines.TPM1
        else:
            return tpm_defines.TPM2

    def read_tpm_interface(self) -> str:
        if self.version == tpm_defines.TPM1:
            return tpm_defines.TPM_FIFO_LEGACY
        tpm_interface_id = self.cs.mem.read_physical_mem_dword(tpm_defines.TPM2_INTERFACE_ADDR)
        tpm_interface_id = (int(tpm_interface_id) & 0xF)
        if tpm_interface_id == 0x0:
            return tpm_defines.TPM_CRB
        elif tpm_interface_id == 0x1:
            return tpm_defines.TPM_FIFO
        elif tpm_interface_id == 0xF:
            return tpm_defines.TPM_FIFO_LEGACY
        else:
            raise RuntimeError('No TPM interface recognized')

    def init_tpm(self) -> tpm_interface.TPM_BASE:
        if self.version == tpm_defines.TPM1:
            return tpm_interface.TPM1(self.cs)
        elif self.version == tpm_defines.TPM2:
            return tpm_interface.TPM2(self.cs, self.interface)
        else:
            raise RuntimeError('Invalid combination of TPM version and interface')
    
    def command(self, commandName: str, locality: str, command_argv: str) -> None:
        self.tpm.command(commandName, locality, command_argv)

    def send_command(self, Locality: int, command: bytes, size: int) -> None:
        self.tpm.send_command(Locality, command, size)

    def read_response(self, Locality: int):
        self.tpm.read_response(Locality)

    def dump_all(self, locality: str) -> None:
        for reg in tpm_defines.list_of_registers:
            self.dump_register(reg, locality)
            
    def log_register_header(self, register_name: str, locality: str) -> None:
        num_spaces = 32 + (-len(register_name) // 2)  # ceiling division
        self.logger.log('=' * 64)
        self.logger.log(f'{" " * num_spaces}{register_name}_{locality}')
        self.logger.log('=' * 64)

    def dump_register(self, register_name: str, locality: str) -> None:
        self.cs.Cfg.REGISTERS[register_name]['address'] = self.cs.Cfg.REGISTERS[register_name]['address'] ^ tpm_defines.LOCALITY[locality]
        register = self.cs.register.read_dict(register_name)

        self.log_register_header(register_name, locality)

        max_field_len = 0
        for field in register['FIELDS']:
            if len(field) > max_field_len:
                max_field_len = len(field)
        for field in register['FIELDS']:
            self.logger.log(f'\t{field}{" " * (max_field_len - len(field))}: {hex(register["FIELDS"][field]["value"])}')

    def identify(self) -> None:
        print(self.tpm)



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

"""
VMM specific functionality
1. Hypervisor hypercall interfaces
2. Second-level Address Translation (SLAT)
3. VirtIO devices
4. ...

"""

import struct

from typing import AnyStr, Dict, List, Optional, Tuple
from chipsec.library.logger import logger, pretty_print_hex_buffer
import chipsec.hal.pcidb


class VMM:

    def __init__(self, cs):
        self.cs = cs
        self.helper = cs.helper
        self.output = ''
        (self.membuf0_va, self.membuf0_pa) = (0, 0)
        (self.membuf1_va, self.membuf1_pa) = (0, 0)

        chipsec.hal.pcidb.VENDORS[VIRTIO_VID] = VIRTIO_VENDOR_NAME
        chipsec.hal.pcidb.DEVICES[VIRTIO_VID] = VIRTIO_DEVICES

    def __del__(self):
        if self.membuf0_va != 0:
            (self.membuf0_va, self.membuf0_pa) = (0, 0)
            (self.membuf1_va, self.membuf1_pa) = (0, 0)

    def init(self) -> None:
        (self.membuf0_va, self.membuf0_pa) = self.cs.mem.alloc_physical_mem(0x2000, 0xFFFFFFFFFFFFFFFF)
        (self.membuf1_va, self.membuf1_pa) = (self.membuf0_va + 0x1000, self.membuf0_pa + 0x1000)
        if self.membuf0_va == 0:
            logger().log("[vmm] Could not allocate memory!")
            raise Exception("[vmm] Could not allocate memory!")

    # Generic hypercall interface

    def hypercall(self, rax: int, rbx: int, rcx: int, rdx: int, rdi: int, rsi: int, r8: int = 0, r9: int = 0, r10: int = 0, r11: int = 0, xmm_buffer: int = 0) -> int:
        return self.helper.hypercall(rcx, rdx, r8, r9, r10, r11, rax, rbx, rdi, rsi, xmm_buffer)

    # Hypervisor-specific hypercall interfaces

    def hypercall64_five_args(self, vector: int, arg1: int = 0, arg2: int = 0, arg3: int = 0, arg4: int = 0, arg5: int = 0) -> int:
        return self.helper.hypercall(0, arg3, arg5, 0, arg4, 0, vector, 0, arg1, arg2)

    def hypercall64_memory_based(self, hypervisor_input_value: int, parameters: AnyStr, size: int = 0) -> int:
        self.cs.mem.write_physical_mem(self.membuf0_pa, len(parameters[:0x1000]), parameters[:0x1000])
        regs = self.helper.hypercall(hypervisor_input_value & ~0x00010000, self.membuf0_pa, self.membuf1_pa)
        self.output = self.helper.read_phys_mem(self.membuf1_pa, size) if size > 0 else ''
        return regs

    def hypercall64_fast(self, hypervisor_input_value: int, param0: int = 0, param1: int = 0) -> int:
        return self.helper.hypercall(hypervisor_input_value | 0x00010000, param0, param1)

    def hypercall64_extended_fast(self, hypervisor_input_value: int, parameter_block: bytes) -> int:
        (param0, param1, xmm_regs) = struct.unpack('<QQ96s', parameter_block)
        self.cs.mem.write_physical_mem(self.membuf0_pa, 0x60, xmm_regs)
        return self.helper.hypercall(hypervisor_input_value | 0x00010000, param0, param1, 0, 0, 0, 0, 0, 0, 0, self.membuf0_va)

    #
    # Dump EPT page tables at specified physical base (EPT pointer)
    #
    def dump_EPT_page_tables(self, eptp: str, pt_fname: Optional[str] = None) -> None:
        _orig_logname = logger().LOG_FILE_NAME
        paging_ept = chipsec.hal.paging.c_extended_page_tables(self.cs)
        logger().log_hal(f'[vmm] Dumping EPT paging hierarchy at EPTP 0x{eptp:08X}...')
        if pt_fname is None:
            pt_fname = (f'ept_{eptp:08X}')
        logger().set_log_file(pt_fname, False)
        paging_ept.read_pt_and_show_status(pt_fname, 'EPT', eptp)
        logger().set_log_file(_orig_logname, False)
        if paging_ept.failure:
            logger().log_error('Could not dump EPT page tables')


################################################################################
#
# VirtIO functions
#
################################################################################

VIRTIO_VID: int = 0x1AF4
VIRTIO_VENDOR_NAME: str = 'Red Hat, Inc.'
VIRTIO_VENDORS: List[int] = [VIRTIO_VID]
VIRTIO_DEVICES: Dict[int, str] = {
    0x1000: 'VirtIO Network',
    0x1001: 'VirtIO Block',
    0x1002: 'VirtIO Baloon',
    0x1003: 'VirtIO Console',
    0x1004: 'VirtIO SCSI',
    0x1005: 'VirtIO RNG',
    0x1009: 'VirtIO filesystem',
    0x1041: 'VirtIO network (1.0)',
    0x1042: 'VirtIO block (1.0)',
    0x1043: 'VirtIO console (1.0)',
    0x1044: 'VirtIO RNG (1.0)',
    0x1045: 'VirtIO memory balloon (1.0)',
    0x1046: 'VirtIO SCSI (1.0)',
    0x1049: 'VirtIO filesystem (1.0)',
    0x1050: 'VirtIO GPU (1.0)',
    0x1052: 'VirtIO input (1.0)',
    0x1110: 'VirtIO Inter-VM shared memory'
}


def get_virtio_devices(devices: List[Tuple[int, int, int, int, int]]) -> List[Tuple[int, int, int, int, int]]:
    virtio_devices = []
    for (b, d, f, vid, did) in devices:
        if vid in VIRTIO_VENDORS:
            virtio_devices.append((b, d, f, vid, did))
    return virtio_devices


class VirtIO_Device:

    def __init__(self, cs, b, d, f):
        self.cs = cs
        self.bus = b
        self.dev = d
        self.fun = f

    def dump_device(self) -> None:
        logger().log(f"\n[vmm] VirtIO device {self.bus:02X}:{self.dev:02X}.{self.fun:01X}")
        dev_cfg = self.cs.pci.dump_pci_config(self.bus, self.dev, self.fun)
        pretty_print_hex_buffer(dev_cfg)
        bars = self.cs.pci.get_device_bars(self.bus, self.dev, self.fun)
        for (bar, isMMIO, _, _, _, size) in bars:
            if isMMIO:
                self.cs.mmio.dump_MMIO(bar, size)
            else:
                self.cs.io.dump_IO(bar, size, 4)
